#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The array_analysis test suite.
"""
import numpy as np
import pytest

from obspy.signal.array_analysis import array_rotation_strain, get_geometry


class TestArray():
    """
    Test cases for array_analysis functions.
    """
    @pytest.fixture(autouse=True, scope="function")
    def setup(self):
        self.array_coords = np.array([[0.0, 0.0, 0.0],
                                     [-5.0, 7.0, 0.0],
                                     [5.0, 7.0, 0.0],
                                     [10.0, 0.0, 0.0],
                                     [5.0, -7.0, 0.0],
                                     [-5.0, -7.0, 0.0],
                                     [-10.0, 0.0, 0.0]])
        self.subarray = np.array([0, 1, 2, 3, 4, 5, 6])
        self.ts1 = np.empty((1000, 7))
        self.ts2 = np.empty((1000, 7))
        self.ts3 = np.empty((1000, 7))
        self.ts1.fill(np.nan)
        self.ts2.fill(np.nan)
        self.ts3.fill(np.nan)
        self.sigmau = 0.0001
        self.Vp = 1.93
        self.Vs = 0.326

    def test_array_rotation(self):
        # tests function array_rotation_strain with synthetic data with pure
        # rotation and no strain
        array_coords = self.array_coords
        subarray = self.subarray
        ts1 = self.ts1
        ts2 = self.ts2
        ts3 = self.ts3
        sigmau = self.sigmau
        vp = self.Vp
        vs = self.Vs

        rotx = 0.00001 * np.exp(-1 * np.square(np.linspace(-2, 2, 1000))) * \
            np.sin(np.linspace(-30 * np.pi, 30 * np.pi, 1000))
        roty = 0.00001 * np.exp(-1 * np.square(np.linspace(-2, 2, 1000))) * \
            np.sin(np.linspace(-20 * np.pi, 20 * np.pi, 1000))
        rotz = 0.00001 * np.exp(-1 * np.square(np.linspace(-2, 2, 1000))) * \
            np.sin(np.linspace(-10 * np.pi, 10 * np.pi, 1000))

        for stat in range(7):
            for t in range(1000):
                ts1[t, stat] = -1. * array_coords[stat, 1] * rotz[t]
                ts2[t, stat] = array_coords[stat, 0] * rotz[t]
                ts3[t, stat] = array_coords[stat, 1] * rotx[t] - \
                    array_coords[stat, 0] * roty[t]

        out = array_rotation_strain(subarray, ts1, ts2, ts3, vp, vs,
                                    array_coords, sigmau)

        np.testing.assert_array_almost_equal(rotx, out['ts_w1'], decimal=12)
        np.testing.assert_array_almost_equal(roty, out['ts_w2'], decimal=12)
        np.testing.assert_array_almost_equal(rotz, out['ts_w3'], decimal=12)
        np.testing.assert_array_almost_equal(np.zeros(1000), out['ts_s'],
                                             decimal=15)
        np.testing.assert_array_almost_equal(np.zeros(1000), out['ts_d'],
                                             decimal=15)
        np.testing.assert_array_almost_equal(np.zeros(1000), out['ts_m'],
                                             decimal=12)

    def test_array_dilation(self):
        # tests function array_rotation_strain with synthetic data with pure
        # dilation and no rotation or shear strain
        array_coords = self.array_coords
        subarray = self.subarray
        ts1 = self.ts1
        ts2 = self.ts2
        ts3 = self.ts3
        sigmau = self.sigmau
        vp = self.Vp
        vs = self.Vs

        eta = 1 - 2 * vs ** 2 / vp ** 2

        dilation = .00001 * np.exp(
            -1 * np.square(np.linspace(-2, 2, 1000))) * \
            np.sin(np.linspace(-40 * np.pi, 40 * np.pi, 1000))

        for stat in range(7):
            for t in range(1000):
                ts1[t, stat] = array_coords[stat, 0] * dilation[t]
                ts2[t, stat] = array_coords[stat, 1] * dilation[t]
                ts3[t, stat] = array_coords[stat, 2] * dilation[t]

        out = array_rotation_strain(subarray, ts1, ts2, ts3, vp, vs,
                                    array_coords, sigmau)

        # remember free surface boundary conditions!
        # see Spudich et al, 1995, (A2)
        np.testing.assert_array_almost_equal(dilation * (2 - 2 * eta),
                                             out['ts_d'], decimal=12)
        np.testing.assert_array_almost_equal(dilation * 2, out['ts_dh'],
                                             decimal=12)
        np.testing.assert_array_almost_equal(
            abs(dilation * .5 * (1 + 2 * eta)), out['ts_s'], decimal=12)
        np.testing.assert_array_almost_equal(np.zeros(1000), out['ts_sh'],
                                             decimal=12)
        np.testing.assert_array_almost_equal(np.zeros(1000), out['ts_w1'],
                                             decimal=15)
        np.testing.assert_array_almost_equal(np.zeros(1000), out['ts_w2'],
                                             decimal=15)
        np.testing.assert_array_almost_equal(np.zeros(1000), out['ts_w3'],
                                             decimal=15)
        np.testing.assert_array_almost_equal(np.zeros(1000), out['ts_m'],
                                             decimal=12)

    def test_array_horizontal_shear(self):
        # tests function array_rotation_strain with synthetic data with pure
        # horizontal shear strain, no rotation or dilation
        array_coords = self.array_coords
        subarray = self.subarray
        ts1 = self.ts1
        ts2 = self.ts2
        sigmau = self.sigmau
        vp = self.Vp
        vs = self.Vs

        shear_strainh = .00001 * np.exp(
            -1 * np.square(np.linspace(-2, 2, 1000))) * \
            np.sin(np.linspace(-10 * np.pi, 10 * np.pi, 1000))

        ts3 = np.zeros((1000, 7))

        for stat in range(7):
            for t in range(1000):
                ts1[t, stat] = array_coords[stat, 1] * shear_strainh[t]
                ts2[t, stat] = array_coords[stat, 0] * shear_strainh[t]

        out = array_rotation_strain(subarray, ts1, ts2, ts3, vp, vs,
                                    array_coords, sigmau)

        np.testing.assert_array_almost_equal(np.zeros(1000), out['ts_d'],
                                             decimal=12)
        np.testing.assert_array_almost_equal(np.zeros(1000), out['ts_dh'],
                                             decimal=12)
        np.testing.assert_array_almost_equal(abs(shear_strainh), out['ts_s'],
                                             decimal=12)
        np.testing.assert_array_almost_equal(abs(shear_strainh), out['ts_sh'],
                                             decimal=12)
        np.testing.assert_array_almost_equal(np.zeros(1000), out['ts_w1'],
                                             decimal=12)
        np.testing.assert_array_almost_equal(np.zeros(1000), out['ts_w2'],
                                             decimal=12)
        np.testing.assert_array_almost_equal(np.zeros(1000), out['ts_w3'],
                                             decimal=12)
        np.testing.assert_array_almost_equal(np.zeros(1000), out['ts_m'],
                                             decimal=12)

    def test_get_geometry(self):
        """
        test get_geometry() in array_analysis.py
        """
        ll = np.array([[24.5797167, 121.4842444, 385.106],
                       [24.5797611, 121.4842333, 384.893],
                       [24.5796694, 121.4842556, 385.106]])

        la = get_geometry(ll)

        np.testing.assert_almost_equal(la[:, 0].sum(), 0., decimal=8)
        np.testing.assert_almost_equal(la[:, 1].sum(), 0., decimal=8)
        np.testing.assert_almost_equal(la[:, 2].sum(), 0., decimal=8)

        ll = np.array([[10., 10., 10.],
                       [0., 5., 5.],
                       [0., 0., 0.]])

        la = get_geometry(ll, coordsys='xy')

        np.testing.assert_almost_equal(la[:, 0].sum(), 0., decimal=8)
        np.testing.assert_almost_equal(la[:, 1].sum(), 0., decimal=8)
        np.testing.assert_almost_equal(la[:, 2].sum(), 0., decimal=8)



# -*- coding: utf-8 -*-
"""
Routing client for the EIDAWS routing service.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import collections

from ..client import get_bulk_string
from ..header import FDSNNoDataException
from .routing_client import (
    BaseRoutingClient, _assert_attach_response_not_in_kwargs,
    _assert_filename_not_in_kwargs, _assert_format_not_in_kwargs)


class EIDAWSRoutingClient(BaseRoutingClient):
    """
    Routing client for the EIDAWS routing service.

    http://www.orfeus-eu.org/data/eida/webservices/routing/

    For waveform queries it will first launch a station query, get the
    station information at each data center with additional constraints
    (e.g. latitude/longitude/...) and use that information for the final
    waveform query. This means that with ObsPy the EIDA routing client
    behaves very similar to the EarthScope (former IRIS) federator routing
    client.
    """
    def __init__(self, url="http://www.orfeus-eu.org/eidaws/routing/1",
                 include_providers=None, exclude_providers=None,
                 debug=False, timeout=120, **kwargs):
        """
        Initialize an EIDAWS router client.

        All parameters except ``url`` are passed on to the
        :class:`~obspy.clients.fdsn.routing.routing_client.BaseRoutingClient`
        parent class

        :param url: The URL of the routing service.
        :type url: str
        """
        BaseRoutingClient.__init__(self, debug=debug, timeout=timeout,
                                   include_providers=include_providers,
                                   exclude_providers=exclude_providers,
                                   **kwargs)
        self._url = url

    @_assert_filename_not_in_kwargs
    @_assert_attach_response_not_in_kwargs
    def get_waveforms_bulk(self, bulk, **kwargs):
        """
        Get waveforms from multiple data centers.

        Arguments are the same as in
        :meth:`obspy.clients.fdsn.client.Client.get_waveforms_bulk()`.
        Any additional ``**kwargs`` are passed on to each individual service's
        dataselect service if the service supports them (otherwise they are
        silently ignored for that particular fdsnws endpoint).

        The ``filename`` and ``attach_response`` parameters of the single
        provider FDSN client are not supported.

        This can route on a number of different parameters, please see the
        web site of the `EIDAWS Routing Service
        <http://www.orfeus-eu.org/data/eida/webservices/routing/>`_
        for details.
        """
        # Multi-step procedure - first get the stations to be able to use
        # more query parameters - and then construct the waveform string
        # from it.
        #
        # This has to be done for each time interval - otherwise it will get
        # a lot more complicated. I guess in most cases people will use bulk
        # requests for the same time span so it should be fine.

        # Group by time interval - utilize the existing get_bulk_string()
        # method to not have to deal with various different inputs.
        _tmp_bulk_str = get_bulk_string(bulk, {})
        if hasattr(_tmp_bulk_str, "decode"):
            _tmp_bulk_str = _tmp_bulk_str.decode()

        # Parse and split.
        bulk_per_time_interval = collections.defaultdict(list)
        for line in _tmp_bulk_str.splitlines():
            # Cannot really happen - just a safety measure.
            if not line:  # pragma: no cover
                continue
            item = line.split()
            bulk_per_time_interval[(item[-2], item[-1])].append(item)

        # Build up the new bulk string for each found time interval by
        # querying the station services.
        new_bulk = []
        for t, _b in bulk_per_time_interval.items():
            # channel level and text to keep it fast.
            inv = self.get_stations_bulk(_b, level="channel", **kwargs)
            for c in sorted(set(inv.get_contents()["channels"])):
                new_bulk.append(c.split("."))
                new_bulk[-1].extend(t)

        # no available data, show appropriate error message and raise
        if not new_bulk:
            msg = ('No data available for request (requested time window '
                   'might be out of bounds of valid station epochs).')
            raise FDSNNoDataException(msg)

        # Finally get the waveforms by getting the routes and downloading
        # everytyhing. Don't directly pass in the initializer as the order
        # would not be guaranteed across all Python version.
        arguments = collections.OrderedDict()
        arguments["service"] = "dataselect"
        arguments["format"] = "post"

        bulk_str = get_bulk_string(new_bulk, arguments)
        r = self._download(self._url + "/query", data=bulk_str,
                           content_type='text/plain')
        split = self._split_routing_response(
            r.content.decode() if hasattr(r.content, "decode") else r.content)
        return self._download_waveforms(split, **kwargs)

    @_assert_filename_not_in_kwargs
    def get_stations(self, **kwargs):
        """
        Get stations from multiple data centers.

        Only the ``network``, ``station``, ``location``, ``channel``,
        ``starttime``, and ``endtime`` parameters are used for the actual
        routing. These and all other arguments are then just passed on to each
        single fdsnws station implementation.

        Arguments are the same as in
        :meth:`obspy.clients.fdsn.client.Client.get_stations()`.
        Any additional ``**kwargs`` are passed on to each individual service's
        station service if the service supports them (otherwise they are
        silently ignored for that particular fdsnws endpoint).

        The ``filename`` parameter of the single provider FDSN client is not
        supported for practical reasons.

        This can route on a number of different parameters, please see the
        web site of the `EIDAWS Routing Service
        <http://www.orfeus-eu.org/data/eida/webservices/routing/>`_
        for details.
        """
        return super(EIDAWSRoutingClient, self).get_stations(**kwargs)

    @_assert_format_not_in_kwargs
    @_assert_filename_not_in_kwargs
    def get_stations_bulk(self, bulk, **kwargs):
        """
        Bulk station download from multiple stations.

        Arguments are the same as in
        :meth:`obspy.clients.fdsn.client.Client.get_stations_bulk()`.
        Any additional ``**kwargs`` are passed on to each individual service's
        station service if the service supports them (otherwise they are
        silently ignored for that particular fdsnws endpoint).

        The ``filename`` and ``format`` parameters of the single provider FDSN
        client are not supported for practical reasons.

        This can route on a number of different parameters, please see the
        web site of the `EIDAWS Routing Service
        <http://www.orfeus-eu.org/data/eida/webservices/routing/>`_
        for details.
        """
        arguments = collections.OrderedDict()
        arguments["service"] = "station"
        arguments["format"] = "post"
        arguments["alternative"] = "false"
        bulk_str = get_bulk_string(bulk, arguments)
        r = self._download(self._url + "/query", data=bulk_str,
                           content_type='text/plain')
        split = self._split_routing_response(
            r.content.decode() if hasattr(r.content, "decode") else r.content)
        return self._download_stations(split, **kwargs)

    @staticmethod
    def _split_routing_response(data):
        """
        Splits the routing responses per data center for the EIDAWS output.

        Returns a dictionary with the keys being the root URLs of the fdsnws
        endpoints and the values the data payloads for that endpoint.

        :param data: The return value from the EIDAWS routing service.
        """
        split = collections.defaultdict(list)
        current_key = None
        for line in data.splitlines():
            line = line.strip()
            if not line:
                continue
            if "http" in line and "fdsnws" in line:
                current_key = line[:line.rfind("/fdsnws")]
                continue
            split[current_key].append(line)

        return {k: "\n".join(v) for k, v in split.items()}



# -*- coding: utf-8 -*-
"""
USAGE: make_c_coverage.py output_dir
"""

import os
import shutil
import subprocess
import sys
from fnmatch import fnmatch

from lxml.html import Element, fromstring, tostring


try:
    target_dir = sys.argv[1]
except IndexError:
    raise SystemExit(__doc__)
obspy_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                         os.path.pardir, os.path.pardir))
build_dir = os.path.join(obspy_dir, 'build')


def cleanup(wildcard='*.o'):
    for root, dirs, files in os.walk(obspy_dir):
        for f in files:
            if fnmatch(f, wildcard):
                os.unlink(os.path.join(root, f))


# cleanup to force rebuild, python setup.py clean --all develop does not
# force a rebuild of all files, therefore manually cleaning up here
shutil.rmtree(target_dir, ignore_errors=True)
shutil.rmtree(build_dir, ignore_errors=True)
cleanup('*.o')
cleanup('*.so')

# GENERATE COVERAGE
os.environ['CFLAGS'] = "-O0 -fprofile-arcs -ftest-coverage"
os.environ['FFLAGS'] = "-O0 -fprofile-arcs -ftest-coverage -fPIC"
os.environ['OBSPY_C_COVERAGE'] = "True"
subprocess.call([sys.executable, 'setup.py', '-v', 'develop'], cwd=obspy_dir)
subprocess.call(['obspy-runtests', '-d'], cwd=obspy_dir)


# FIND ALL COVERAGE PROFILE STATISTICS
profs = []
for root, dirs, files in os.walk(build_dir):
    profs.extend(os.path.abspath(os.path.join(root, i))
                 for i in files if fnmatch(i, '*.gcda'))

# GENERATE REPORTS WITH GCOV
cov = []
for gcda in profs:
    source = gcda[gcda.rfind('obspy' + os.path.sep):].replace('gcda', 'c')
    if not os.path.exists(os.path.join(obspy_dir, source)):
        source = source.replace('.c', '.f')

    output = subprocess.check_output(['gcov', '--object-file', gcda, source],
                                     cwd=obspy_dir)
    output = output.decode().splitlines()

    filename = output[0].strip().split()[1].strip("'")
    perc = float(output[1].split(':')[1].split('%')[0])
    gcov = output[2].strip().split()[1].strip("'")

    # move generated gcov to coverage folder
    new_dir = os.path.join(target_dir, os.path.dirname(source))
    try:
        os.makedirs(new_dir)
    except OSError:
        pass
    os.rename(os.path.join(obspy_dir, gcov), os.path.join(new_dir, gcov))
    cov.append((filename, os.path.join(new_dir, gcov), perc))


# GENERATE HTML
page = fromstring("<html><table></table></html>")
table = page.xpath('.//table')[0]
for name, gcov, perc in cov:
    td1, td2 = Element('td'), Element('td')
    gcov = gcov.replace(target_dir, './')
    a = Element('a', attrib={'href': gcov})
    a.text = name
    td1.append(a)
    td2.text = "%6.2f%%" % perc
    tr = Element('tr')
    tr.extend([td1, td2])
    table.append(tr)
with open(os.path.join(target_dir, 'index.html'), 'wb') as fp:
    fp.write(tostring(page))

cleanup('*.o')



# -*- coding: utf-8 -*-
import math
import warnings
import numpy as np

from obspy.geodetics import (calc_vincenty_inverse, degrees2kilometers,
                             gps2dist_azimuth, inside_geobounds,
                             kilometer2degrees, locations2degrees)
from obspy.geodetics.base import HAS_GEOGRAPHICLIB
from obspy.core import AttribDict
import pytest


def dms2dec(degs, mins, secs):
    """Converts angle given in degrees, mins and secs to decimal degrees"""
    return (degs + mins / 60.0 + secs / 3600.0)


class TestUtilGeodetics:
    """
    Test suite for obspy.core.util.geodetics
    """
    def test_calc_vincenty_inverse(self):
        """
        Tests for the Vincenty's Inverse formulae.
        """
        # the following will raise StopIteration exceptions because of two
        # nearly antipodal points
        with pytest.raises(StopIteration):
            calc_vincenty_inverse(
                15.26804251, 2.93007342, -14.80522806, -177.2299081,
            )
        with pytest.raises(StopIteration):
            calc_vincenty_inverse(
                27.3562106, 72.2382356, -27.55995499, -107.78571981,
            )
        with pytest.raises(StopIteration):
            calc_vincenty_inverse(
                27.4675551, 17.28133229, -27.65771704, -162.65420626,
            )
        with pytest.raises(StopIteration):
            calc_vincenty_inverse(
                27.4675551, 17.28133229, -27.65771704, -162.65420626,
            )
        # working examples
        res = calc_vincenty_inverse(0, 0.2, 0, 20)
        assert round(abs(res[0]-2204125.9174282863), 7) == 0
        assert round(abs(res[1]-90.0), 7) == 0
        assert round(abs(res[2]-270.0), 7) == 0
        res = calc_vincenty_inverse(0, 0, 0, 10)
        assert round(abs(res[0]-1113194.9077920639), 7) == 0
        assert round(abs(res[1]-90.0), 7) == 0
        assert round(abs(res[2]-270.0), 7) == 0
        res = calc_vincenty_inverse(0, 0, 0, 13)
        assert round(abs(res[0]-1447153.3801296828), 7) == 0
        assert round(abs(res[1]-90.0), 7) == 0
        assert round(abs(res[2]-270.0), 7) == 0
        res = calc_vincenty_inverse(0, 0, 0, 17)
        assert round(abs(res[0]-1892431.3432465086), 7) == 0
        assert round(abs(res[1]-90.0), 7) == 0
        assert round(abs(res[2]-270.0), 7) == 0
        # out of bounds
        with pytest.raises(ValueError):
            calc_vincenty_inverse(91, 0, 0, 0)
        with pytest.raises(ValueError):
            calc_vincenty_inverse(-91, 0, 0, 0)
        with pytest.raises(ValueError):
            calc_vincenty_inverse(0, 0, 91, 0)
        with pytest.raises(ValueError):
            calc_vincenty_inverse(0, 0, -91, 0)

    @pytest.mark.skipif(
        not HAS_GEOGRAPHICLIB, reason='Module geographiclib is not installed'
    )
    def test_gps_2_dist_azimuth_with_geographiclib(self):
        """
        Testing gps2dist_azimuth function using the module geographiclib.
        """
        # nearly antipodal points
        result = gps2dist_azimuth(15.26804251, 2.93007342, -14.80522806,
                                  -177.2299081)
        assert round(abs(result[0]-19951425.048688546), 7) == 0
        assert round(abs(result[1]-8.65553241932755), 7) == 0
        assert round(abs(result[2]-351.36325485132306), 7) == 0
        # out of bounds
        with pytest.raises(ValueError):
            gps2dist_azimuth(91, 0, 0, 0)
        with pytest.raises(ValueError):
            gps2dist_azimuth(-91, 0, 0, 0)
        with pytest.raises(ValueError):
            gps2dist_azimuth(0, 0, 91, 0)
        with pytest.raises(ValueError):
            gps2dist_azimuth(0, 0, -91, 0)

    def test_calc_vincenty_inverse_2(self):
        """
        Test calc_vincenty_inverse() method with test data from Geocentric
        Datum of Australia.
        see https://www.icsm.gov.au/publications/gda2020-technical-manual-v16
        """
        # test data:
        # Point 1: Flinders Peak, Point 2: Buninyong
        lat1 = -(37 + (57 / 60.) + (3.72030 / 3600.))
        lon1 = 144 + (25 / 60.) + (29.52440 / 3600.)
        lat2 = -(37 + (39 / 60.) + (10.15610 / 3600.))
        lon2 = 143 + (55 / 60.) + (35.38390 / 3600.)
        dist = 54972.271
        alpha12 = 306 + (52 / 60.) + (5.37 / 3600.)
        alpha21 = 127 + (10 / 60.) + (25.07 / 3600.)

        # calculate result
        calc_dist, calc_alpha12, calc_alpha21 = calc_vincenty_inverse(
            lat1, lon1, lat2, lon2)

        # calculate deviations from test data
        dist_err_rel = abs(dist - calc_dist) / dist
        alpha12_err = abs(alpha12 - calc_alpha12)
        alpha21_err = abs(alpha21 - calc_alpha21)

        assert dist_err_rel < 1.0e-5
        assert alpha12_err < 1.0e-5
        assert alpha21_err < 1.0e-5

        # calculate result with +- 360 for lon values
        dist, alpha12, alpha21 = calc_vincenty_inverse(
            lat1, lon1 + 360, lat2, lon2 - 720)
        assert round(abs(dist-calc_dist), 7) == 0
        assert round(abs(alpha12-calc_alpha12), 7) == 0
        assert round(abs(alpha21-calc_alpha21), 7) == 0

    def test_calc_vincenty_inverse_tabulated(self):
        """ Tabulated results for Vincenty Inverse

        Table II of Vincenty's paper (T. Vincenty 1975, "Direct and inverse
        solutions of geodesics on the ellipsoid with application of nested
        equations" Survey Review XXII pp.88-93) has five test examples for
        the forward and inverse problem (with results rounded to 0.00001
        seconds of arc and 1 mm). The inverse versions of these are implemented
        here. Note the non-standard (old) ellipsoid usage. Here we test that
        we match these examples for the inverse problem. """
        # Row "A"
        # NB: for this case there seems to be a typo in
        #     the tabulated data. Tabulated data is commented
        #     out and values from geographiclib are used in their place
        # dist = 14110526.170
        dist = 14039003.954192352
        # azi1 = dms2dec(96.0, 36.0, 8.79960)
        azi1 = 95.88145755849257
        # azi2 = dms2dec(137.0, 52.0, 22.01454)
        azi2 = 138.30481836546775
        bazi = azi2 + 180.0
        lat1 = dms2dec(55.0, 45.0, 0.0)
        lat2 = dms2dec(-33.0, 26.0, 0.0)
        lon2 = dms2dec(108.0, 13.0, 0.0)
        a = 6377397.155
        f = 1.0 / 299.1528128
        calc_dist, calc_azi1, calc_bazi = calc_vincenty_inverse(
            lat1, 0.0, lat2, lon2, a, f)
        assert round(abs(dist-calc_dist), 2) == 0
        assert round(abs(azi1-calc_azi1), 5) == 0
        assert round(abs(bazi-calc_bazi), 5) == 0

        # Row "B"
        dist = 4085966.703
        azi1 = dms2dec(95.0, 27.0, 59.63089)
        azi2 = dms2dec(118, 5.0, 58.96161)
        bazi = azi2 + 180.0
        lat1 = dms2dec(37.0, 19.0, 54.95367)
        lat2 = dms2dec(26.0, 7.0, 42.83946)
        lon2 = dms2dec(41.0, 28.0, 35.50729)
        a = 6378388.000
        f = 1.0 / 297.0
        calc_dist, calc_azi1, calc_bazi = calc_vincenty_inverse(
            lat1, 0.0, lat2, lon2, a, f)
        assert round(abs(dist-calc_dist), 2) == 0
        assert round(abs(azi1-calc_azi1), 5) == 0
        assert round(abs(bazi-calc_bazi), 5) == 0

        # Row "C"
        dist = 8084823.839
        azi1 = dms2dec(15.0, 44.0, 23.74850)
        azi2 = dms2dec(144.0, 55.0, 39.92147)
        bazi = azi2 + 180.0
        lat1 = dms2dec(35.0, 16.0, 11.24862)
        lat2 = dms2dec(67.0, 22.0, 14.77638)
        lon2 = dms2dec(137.0, 47.0, 28.31435)
        a = 6378388.000
        f = 1.0 / 297.0
        calc_dist, calc_azi1, calc_bazi = calc_vincenty_inverse(
            lat1, 0.0, lat2, lon2, a, f)
        assert round(abs(dist-calc_dist), 2) == 0
        assert round(abs(azi1-calc_azi1), 5) == 0
        assert round(abs(bazi-calc_bazi), 5) == 0

    @pytest.mark.skipif(
        HAS_GEOGRAPHICLIB,
        reason='Geographiclib installed, not using calc_vincenty_inverse'
    )
    def test_gps_2_dist_azimuth_bug150(self):
        """
        Test case for #150: UserWarning will be only raised if geographiclib is
        not installed.
        """
        # this raises UserWarning
        with warnings.catch_warnings(record=True):
            warnings.simplefilter('error', UserWarning)
            with pytest.raises(UserWarning):
                gps2dist_azimuth(0, 0, 0, 180)

    def test_kilometer2degrees(self):
        """
        Simple test of the convenience function.
        """
        # Test if it works.
        assert kilometer2degrees(111.19492664455873, radius=6371) == 1.0
        # Test if setting the radius actually does something. Round to avoid
        # some precision problems on different machines.
        assert round(kilometer2degrees(111.19492664455873,
                     radius=6381), 5) == round(0.99843284751606332, 5)

    def test_degrees2kilometers(self):
        """
        """
        # Test if it works.
        assert degrees2kilometers(1.0, radius=6371) == 111.19492664455873
        # Test if setting the radius actually does something. Round to avoid
        # some precision problems on different machines.
        assert round(degrees2kilometers(1.0, radius=6381), 5) == \
               round(111.36945956975816, 5)

    def test_locations2degrees(self):
        """
        Test the location 2 degree conversion.
        """
        # Inline method to avoid messy code.
        def assert_loc(lat1, long1, lat2, long2, approx_distance):
            assert abs(math.radians(locations2degrees(
                lat1, long1, lat2, long2)) * 6371 - approx_distance) <= 20

        # Approximate values from the Great Circle Calculator:
        #   http://williams.best.vwh.net/gccalc.htm

        # Random location.
        assert_loc(36.12, -86.67, 33.94, -118.40, 2893)
        # Test several combinations of quadrants.
        assert_loc(11.11, 22.22, 33.33, 44.44, 3346)
        assert_loc(-11.11, -22.22, -33.33, -44.44, 3346)
        assert_loc(11.11, 22.22, -33.33, -44.44, 8596)
        assert_loc(-11.11, -22.22, 33.33, 44.44, 8596)
        assert_loc(11.11, -22.22, 33.33, -44.44, 3346)
        assert_loc(-11.11, 22.22, 33.33, 44.44, 5454)
        assert_loc(11.11, -22.22, 33.33, 44.44, 7177)
        assert_loc(11.11, 22.22, -33.33, 44.44, 5454)
        assert_loc(11.11, 22.22, 33.33, -44.44, 7177)
        # Test some extreme values.
        assert_loc(90, 0, 0, 0, 10018)
        assert_loc(180, 0, 0, 0, 20004)
        assert_loc(0, 90, 0, 0, 10018)
        assert_loc(0, 180, 0, 0, 20004)
        assert_loc(0, 0, 90, 0, 10018)
        assert_loc(0, 0, 180, 0, 20004)
        assert_loc(0, 0, 0, 90, 10018)
        assert_loc(0, 0, 0, 180, 20004)
        assert_loc(11, 55, 11, 55, 0)

        # test numpy inputs:
        # Inline method to avoid messy code.
        def assert_loc_np(lat1, long1, lat2, long2,
                          approx_distance, expected_output_len):
            loc2deg = locations2degrees(np.array(lat1),
                                        np.array(long1),
                                        np.array(lat2),
                                        np.array(long2))
            assert (np.abs(np.radians(loc2deg) * 6371 -
                    approx_distance) <= 20).all()

            assert np.isscalar(loc2deg) \
                   if expected_output_len == 0 else \
                   len(loc2deg) == expected_output_len

        # Test just with random location (combining scalars and arrays).
        assert_loc_np(36.12, -86.67, 33.94, -118.40, 2893, 0)
        assert_loc_np([36.12, 36.12], -86.67, 33.94, -118.40,
                      2893, 2)
        assert_loc_np(36.12, [-86.67, -86.67], 33.94, -118.40,
                      2893, 2)
        assert_loc_np(36.12, -86.67, [33.94, 33.94], -118.40,
                      2893, 2)
        assert_loc_np(36.12, -86.67, 33.94, [-118.40, -118.40],
                      2893, 2)
        assert_loc_np([36.12, 36.12], [-86.67, -86.67], 33.94, -118.40,
                      2893, 2)
        assert_loc_np([36.12, 36.12], -86.67, [33.94, 33.94], -118.40,
                      2893, 2)
        assert_loc_np([36.12, 36.12], -86.67, 33.94, [-118.40, -118.40],
                      2893, 2)
        assert_loc_np([36.12, 36.12], [-86.67, -86.67], [33.94, 33.94],
                      -118.40, 2893, 2)
        assert_loc_np([36.12, 36.12], -86.67, [33.94, 33.94],
                      [-118.40, -118.40], 2893, 2)
        assert_loc_np(36.12, [-86.67, -86.67], [33.94, 33.94],
                      [-118.40, -118.40], 2893, 2)
        assert_loc_np([36.12, 36.12], [-86.67, -86.67], [33.94, 33.94],
                      [-118.40, -118.40], 2893, 2)

        # test numpy broadcasting (bad shapes)
        with pytest.raises(ValueError):
            locations2degrees(1, 2, [3, 4], [5, 6, 7])

    @pytest.mark.skipif(
        not HAS_GEOGRAPHICLIB,
        reason='Module geographiclib is not installed',
    )
    def test_issue_375(self):
        """
        Test for #375.
        """
        _, azim, bazim = gps2dist_azimuth(50, 10, 50 + 1, 10 + 1)
        assert round(azim, 0) == 32
        assert round(bazim, 0) == 213
        _, azim, bazim = gps2dist_azimuth(50, 10, 50 + 1, 10 - 1)
        assert round(azim, 0) == 328
        assert round(bazim, 0) == 147
        _, azim, bazim = gps2dist_azimuth(50, 10, 50 - 1, 10 + 1)
        assert round(azim, 0) == 147
        assert round(bazim, 0) == 327
        _, azim, bazim = gps2dist_azimuth(50, 10, 50 - 1, 10 - 1)
        assert round(azim, 0) == 213
        assert round(bazim, 0) == 33

    def test_inside_geobounds(self):
        obj = AttribDict()
        obj.latitude = 48.8566
        obj.longitude = 2.3522
        ret = inside_geobounds(obj, minlatitude=48, maxlatitude=49,
                               minlongitude=2, maxlongitude=3)
        assert ret
        ret = inside_geobounds(obj, latitude=48, longitude=2,
                               minradius=1, maxradius=2)
        assert not ret
        # Test for wrapping around longitude +/- 180°
        obj.latitude = -41.2865
        obj.longitude = 174.7762
        ret = inside_geobounds(obj, minlongitude=170, maxlongitude=-170)
        assert ret
        obj.longitude = -175.
        ret = inside_geobounds(obj, minlongitude=170, maxlongitude=-170)
        assert ret
        ret = inside_geobounds(obj, minlongitude=170, maxlongitude=190)
        assert ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
obspy.io.nlloc - NonLinLoc file format support for ObsPy
========================================================

This module provides read/write support for some NonLinLoc file
formats.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)

Example
-------
If NonLinLoc location run was performed using one of the standard projections
available in the NonLinLoc package, it is straightforward to read a NonLinLoc
Hypocenter-Phase file into an ObsPy :class:`~obspy.core.event.Catalog` object:

>>> from obspy import read_events
>>> cat = read_events("/path/to/nlloc.hyp")
>>> print(cat)
1 Event(s) in Catalog:
2006-07-15T17:21:20.195670Z | +51.658,   +7.737

>>> event = cat[0]
>>> print(event)  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
Event:	2006-07-15T17:21:20.195670Z | +51.658,   +7.737
<BLANKLINE>
         resource_id: ResourceIdentifier(id="smi:local/...")
       creation_info: CreationInfo(author='Claudio Satriano',
                                   creation_time=UTCDateTime(2013, 6, 21, ...),
                                   version='NLLoc:v6.02.07')
 preferred_origin_id: ResourceIdentifier(id="smi:local/...")
                ---------
            comments: 2 Elements
               picks: 5 Elements
             origins: 1 Elements

>>> origin = event.origins[0]
>>> print(origin)  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
Origin
         resource_id: ResourceIdentifier(id="smi:local/...")
                time: UTCDateTime(2006, 7, 15, 17, 21, 20, 1956...)
           longitude: 7.736781...
            latitude: 51.657659...
               depth: 1433.5... [confidence_level=68, uncertainty=...]
          depth_type: 'from location'
             quality: OriginQuality(...)
  origin_uncertainty: OriginUncertainty(...)
       creation_info: CreationInfo(author='Claudio Satriano',
                                   creation_time=UTCDateTime(2013, 6, 21, ...),
                                   version='NLLoc:v6.02.07')
                ---------
            comments: 2 Elements
            arrivals: 5 Elements


If, instead, NonLinLoc location run was performed in some custom coordinate
system (as opposed to WGS84 with depth in meters down), we need to set up a
coordinate conversion function to convert from the NonLinLoc location
coordinates `x`, `y`, `z` to longitude, latitude and depth in kilometers.
In the example, the location run was done in Gauß-Krüger zone 4 (EPSG:31468,
but in kilometers for coordinates) and depth in kilometers. So we have to
convert `x` and `y` (of NonLinLoc location run) to meters (as defined in
EPSG:31468) and then we convert to WGS84 (EPSG:4326). The `z` coordinate is
already in kilometers downwards and can be left as is. For the conversion we
use `pyproj`.

>>> import pyproj  # doctest: +SKIP
>>> proj_wgs84 = pyproj.Proj("epsg:4326")  # doctest: +SKIP
>>> proj_gk4 = pyproj.Proj("epsg:31468")  # doctest: +SKIP
>>> def my_conversion(x, y, z):
...     x *= 1e3
...     y *= 1e3
...     x, y = pyproj.transform(proj_gk4, proj_wgs84, x, y)
...     return x, y, z  # doctest: +SKIP

Then, we can load the NonLinLoc Hypocenter-Phase file into an ObsPy
:class:`~obspy.core.event.Catalog` object using
:func:`~obspy.core.event.read_events`, supplying our coordinate mapping
function as `coordinate_converter` kwarg, which will be passed down to the
low-level routine :func:`~obspy.io.nlloc.core.read_nlloc_hyp`.

>>> from obspy import read_events  # doctest: +SKIP
>>> cat = read_events("/path/to/nlloc_custom.hyp",
...                  coordinate_converter=my_conversion)  # doctest: +SKIP
>>> print(cat)  # doctest: +SKIP
1 Event(s) in Catalog:
2010-05-27T16:56:24.612600Z | +48.047,  +11.646

>>> event = cat[0]  # doctest: +SKIP
>>> print(event)  # doctest: +SKIP
Event:  2010-05-27T16:56:24.612600Z | +48.047,  +11.646
<BLANKLINE>
       resource_id: ResourceIdentifier(id="smi:local/...")
     creation_info: CreationInfo(creation_time=..., version='NLLoc:v6.00.0')
    ---------
             picks: 8 Elements
           origins: 1 Elements

>>> origin = event.origins[0]  # doctest: +SKIP
>>> print(origin)  # doctest: +SKIP
Origin
       resource_id: ResourceIdentifier(id="smi:local/...")
              time: UTCDateTime(2010, 5, 27, 16, 56, 24, 612600)
         longitude: 11.64553754...
          latitude: 48.04707051...
             depth: 4579.4... [confidence_level=68, uncertainty=191.6063...]
                ...
     creation_info: CreationInfo(creation_time=UTCDateTime(2014, 10, 17,
16, 30, 8), version='NLLoc:v6.00.0')
    ---------
          comments: 2 Elements
          arrivals: 8 Elements

If NonLinLoc reports the location run as "ABORTED", "IGNORED" or "REJECTED",
the evaluation status of the origin will be set to "rejected" (which otherwise
is unset, i.e. ``None``). Further information might be found in the "NLLOC"
info header line of the hyp file that gets stored in event and origin comments:

>>> cat = read_events("/path/to/nlloc_rejected.hyp", format="NLLOC_HYP")
>>> print(cat[0].origins[0].evaluation_status)
rejected
>>> print(cat[0].comments[1].text)  # doctest: +ELLIPSIS
NLLOC ... "REJECTED" ... max prob location on grid boundary 10, rejecting ...
"""
if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
The cross correlation test suite.
"""
import ctypes as C  # NOQA
import numpy as np
import warnings

import pytest

from obspy import UTCDateTime, read, Trace
from obspy.core.util import AttribDict
from obspy.core.util.libnames import _load_cdll
from obspy.signal.cross_correlation import (
    correlate, correlate_template, correlate_stream_template,
    correlation_detector,
    xcorr_pick_correction, xcorr_3c, xcorr_max,
    _xcorr_padzeros, _xcorr_slice, _find_peaks)
from obspy.signal.trigger import coincidence_trigger


class TestCrossCorrelation:

    """
    Cross correlation test case
    """
    @pytest.fixture(scope='class')
    def state(self):
        """Return test state."""
        out = AttribDict()
        out.a = np.sin(np.linspace(0, 10, 101))
        out.b = 5 * np.roll(out.a, 5)
        out.c = 5 * np.roll(out.a[:81], 5)
        return out

    def test_correlate_normalize_true_false(self):
        a = read()[0].data[500:]
        b = a[10:]
        shift = 100
        cc1 = correlate(a, b, shift, normalize='naive')
        cc2 = correlate(a, b, shift, normalize=True)
        cc3 = correlate(a, b, shift, normalize=None)
        cc4 = correlate(a, b, shift, normalize=False)
        np.testing.assert_allclose(cc1, cc2, rtol=1e-6)
        np.testing.assert_allclose(cc3, cc4, rtol=1e-6)

    def test_srl_xcorr(self):
        """
        Tests if example in ObsPy paper submitted to the Electronic
        Seismologist section of SRL is still working. The test shouldn't be
        changed because the reference gets wrong.
        """
        np.random.seed(815)
        data1 = np.random.randn(1000).astype(np.float32)
        data2 = data1.copy()

        window_len = 100
        corp = np.empty(2 * window_len + 1, dtype=np.float64)

        lib = _load_cdll("signal")
        #
        shift = C.c_int()
        coe_p = C.c_double()
        res = lib.X_corr(data1.ctypes.data_as(C.c_void_p),
                         data2.ctypes.data_as(C.c_void_p),
                         corp.ctypes.data_as(C.c_void_p),
                         window_len, len(data1), len(data2),
                         C.byref(shift), C.byref(coe_p))

        assert 0 == res
        assert round(abs(0.0-shift.value), 7) == 0
        assert round(abs(1.0-coe_p.value), 7) == 0

    def test_xcorr_vs_old_implementation(self, state):
        """
        Test against output of xcorr from ObsPy<1.1
        """
        # Results of xcorr(self.a, self.b, 15, full_xcorr=True)
        # for ObsPy==1.0.2:
        # -5, 0.9651607597888241
        x = [0.53555336, 0.60748967, 0.67493495, 0.73707491, 0.79313226,
             0.84237607, 0.88413089, 0.91778536, 0.94280034, 0.95871645,
             0.96516076, 0.96363672, 0.95043933, 0.92590109, 0.89047807,
             0.84474328, 0.78377236, 0.71629895, 0.64316805, 0.56526677,
             0.48351386, 0.39884904, 0.31222231, 0.22458339, 0.13687123,
             0.05000401, -0.03513057, -0.11768441, -0.19685756, -0.27190599,
             -0.34214866]
        corr_fun = correlate(state.a, state.b, shift=15)
        shift, corr = xcorr_max(corr_fun)
        np.testing.assert_allclose(corr_fun, x)
        assert round(abs(corr-0.96516076), 7) == 0
        assert shift == -5

    def test_correlate_different_length_of_signals(self, state):
        # Signals are aligned around the middle
        cc = correlate(state.a, state.c, 50)
        shift, _ = xcorr_max(cc)
        assert shift == -5 - (len(state.a) - len(state.c)) // 2

    def test_correlate(self):
        # simple test
        a, b = [0, 1], [20, 10]
        cc = correlate(a, b, 1, demean=False, normalize=False)
        shift, value = xcorr_max(cc)
        assert shift == 1
        assert round(abs(value-20.), 7) == 0
        np.testing.assert_allclose(cc, [0., 10., 20.], atol=1e-14)
        # test symetry and different length of a and b
        a, b = [0, 1, 2], [20, 10]
        cc1 = correlate(a, b, 1, demean=False, normalize=False, method='fft')
        cc2 = correlate(a, b, 1, demean=False, normalize=False,
                        method='direct')
        cc3 = correlate(b, a, 1, demean=False, normalize=False, method='fft')
        cc4 = correlate(b, a, 1, demean=False, normalize=False,
                        method='direct')
        shift1, _ = xcorr_max(cc1)
        shift2, _ = xcorr_max(cc2)
        shift3, _ = xcorr_max(cc3)
        shift4, _ = xcorr_max(cc4)
        assert shift1 == 0.5
        assert shift2 == 0.5
        assert shift3 == -0.5
        assert shift4 == -0.5
        np.testing.assert_allclose(cc1, cc2)
        np.testing.assert_allclose(cc3, cc4)
        np.testing.assert_allclose(cc1, cc3[::-1])
        # test sysmetry for method='direct' and len(a) - len(b) - 2 * num > 0
        a, b = [0, 1, 2, 3, 4, 5, 6, 7], [20, 10]
        cc1 = correlate(a, b, 2, method='direct')
        cc2 = correlate(b, a, 2, method='direct')
        np.testing.assert_allclose(cc1, cc2[::-1])

    def test_correlate_different_implementations(self, state):
        """
        Test correct length and different implementations against each other
        """
        xcorrs1 = []
        xcorrs2 = []
        for xcorr_func in (_xcorr_padzeros, _xcorr_slice):
            for method in ('auto', 'fft', 'direct'):
                x = xcorr_func(state.a, state.b, 40, method)
                y = xcorr_func(state.a, state.b[:-1], 40, method)
                assert (len(state.a) - len(state.b)) % 2 == 0
                assert len(x) == 2 * 40 + 1
                assert len(y) == 2 * 40
                xcorrs1.append(x)
                xcorrs2.append(y)
        for x_other in xcorrs1[1:]:
            np.testing.assert_allclose(x_other, xcorrs1[0])
        for x_other in xcorrs2[1:]:
            np.testing.assert_allclose(x_other, xcorrs2[0])

    def test_correlate_extreme_shifts_for_freq_xcorr(self):
        """
        Also test shift=None
        """
        a, b = [1, 2, 3], [1, 2, 3]
        n = len(a) + len(b) - 1
        cc1 = correlate(a, b, 2, method='fft')
        cc2 = correlate(a, b, 3, method='fft')
        cc3 = correlate(a, b, None, method='fft')
        cc4 = correlate(a, b, None, method='direct')
        assert len(cc1) == n
        assert len(cc2) == 2 + n
        assert len(cc3) == n
        assert len(cc4) == n
        a, b = [1, 2, 3], [1, 2]
        n = len(a) + len(b) - 1
        cc1 = correlate(a, b, 2, method='fft')
        cc2 = correlate(a, b, 3, method='fft')
        cc3 = correlate(a, b, None, method='fft')
        cc4 = correlate(a, b, None, method='direct')
        assert len(cc1) == n
        assert len(cc2) == 2 + n
        assert len(cc3) == n
        assert len(cc4) == n

    def test_xcorr_max(self):
        shift, value = xcorr_max((1, 3, -5))
        assert shift == 1
        assert value == -5
        shift, value = xcorr_max((3., -5.), abs_max=False)
        assert shift == -0.5
        assert value == 3.

    def test_xcorr_3c(self):
        st = read()
        st2 = read()
        for tr in st2:
            tr.data = -5 * np.roll(tr.data, 50)
        shift, value, x = xcorr_3c(st, st2, 200, full_xcorr=True)
        assert shift == -50
        assert round(abs(value--0.998), 3) == 0

    def test_xcorr_pick_correction(self, state, testdata):
        """
        Test cross correlation pick correction on a set of two small local
        earthquakes.
        """
        st1 = read(testdata['BW.UH1._.EHZ.D.2010.147.a.slist.gz'])
        st2 = read(testdata['BW.UH1._.EHZ.D.2010.147.b.slist.gz'])

        tr1 = st1.select(component="Z")[0]
        tr2 = st2.select(component="Z")[0]
        tr1_copy = tr1.copy()
        tr2_copy = tr2.copy()
        t1 = UTCDateTime("2010-05-27T16:24:33.315000Z")
        t2 = UTCDateTime("2010-05-27T16:27:30.585000Z")

        dt, coeff = xcorr_pick_correction(t1, tr1, t2, tr2, 0.05, 0.2, 0.1)
        assert round(abs(dt--0.014459080288833711), 7) == 0
        assert round(abs(coeff-0.91542878457939791), 7) == 0
        dt, coeff = xcorr_pick_correction(t2, tr2, t1, tr1, 0.05, 0.2, 0.1)
        assert round(abs(dt-0.014459080288833711), 7) == 0
        assert round(abs(coeff-0.91542878457939791), 7) == 0
        dt, coeff = xcorr_pick_correction(
            t1, tr1, t2, tr2, 0.05, 0.2, 0.1, filter="bandpass",
            filter_options={'freqmin': 1, 'freqmax': 10})
        assert round(abs(dt--0.013025086360067755), 7) == 0
        assert round(abs(coeff-0.98279277273758803), 7) == 0
        assert tr1 == tr1_copy
        assert tr2 == tr2_copy

    def test_xcorr_pick_correction_images(self, state, image_path, testdata):
        """
        Test cross correlation pick correction on a set of two small local
        earthquakes.
        """
        st1 = read(testdata['BW.UH1._.EHZ.D.2010.147.a.slist.gz'])
        st2 = read(testdata['BW.UH1._.EHZ.D.2010.147.b.slist.gz'])

        tr1 = st1.select(component="Z")[0]
        tr2 = st2.select(component="Z")[0]
        t1 = UTCDateTime("2010-05-27T16:24:33.315000Z")
        t2 = UTCDateTime("2010-05-27T16:27:30.585000Z")

        xcorr_pick_correction(
            t1, tr1, t2, tr2, 0.05, 0.2, 0.1, plot=True, filename=image_path)

    def test_correlate_template_eqcorrscan(self):
        """
        Test for moving window correlations with "full" normalisation.

        Comparison result is from EQcorrscan v.0.2.7, using the following:

        from eqcorrscan.utils.correlate import get_array_xcorr
        from obspy import read

        data = read()[0].data
        template = data[400:600]
        data = data[380:620]
        eqcorrscan_func = get_array_xcorr("fftw")
        result = eqcorrscan_func(
            stream=data, templates=template.reshape(1, len(template)),
            pads=[0])[0][0]
        """
        result = [
            -2.24548906e-01, 7.10350871e-02, 2.68642932e-01, 2.75941312e-01,
            1.66854098e-01, 1.66086946e-02, -1.29057273e-01, -1.96172655e-01,
            -1.41613603e-01, -6.83271606e-03, 1.45768464e-01, 2.42143899e-01,
            1.98310092e-01, 2.16377302e-04, -2.41576880e-01, -4.00586188e-01,
            -4.32240069e-01, -2.88735539e-01, 1.26461715e-01, 7.09268868e-01,
            9.99999940e-01, 7.22769439e-01, 1.75955653e-01, -2.46459037e-01,
            -4.34027880e-01, -4.32590246e-01, -2.67131507e-01, -6.78363896e-04,
            2.08171085e-01, 2.32197508e-01, 8.64804164e-02, -1.14158235e-01,
            -2.53621429e-01, -2.62945205e-01, -1.40505865e-01, 3.35594788e-02,
            1.77415669e-01, 2.72263527e-01, 2.81718552e-01, 1.38080209e-01,
            -1.27307668e-01]
        data = read()[0].data
        template = data[400:600]
        data = data[380:620]
        cc = correlate_template(data, template)
        np.testing.assert_allclose(cc, result, atol=1e-7)
        shift, corr = xcorr_max(cc)
        assert round(abs(corr-1.0), 7) == 0
        assert shift == 0

    def test_correlate_template_eqcorrscan_time(self):
        """
        Test full normalization for method='direct'.
        """
        result = [
            -2.24548906e-01, 7.10350871e-02, 2.68642932e-01, 2.75941312e-01,
            1.66854098e-01, 1.66086946e-02, -1.29057273e-01, -1.96172655e-01,
            -1.41613603e-01, -6.83271606e-03, 1.45768464e-01, 2.42143899e-01,
            1.98310092e-01, 2.16377302e-04, -2.41576880e-01, -4.00586188e-01,
            -4.32240069e-01, -2.88735539e-01, 1.26461715e-01, 7.09268868e-01,
            9.99999940e-01, 7.22769439e-01, 1.75955653e-01, -2.46459037e-01,
            -4.34027880e-01, -4.32590246e-01, -2.67131507e-01, -6.78363896e-04,
            2.08171085e-01, 2.32197508e-01, 8.64804164e-02, -1.14158235e-01,
            -2.53621429e-01, -2.62945205e-01, -1.40505865e-01, 3.35594788e-02,
            1.77415669e-01, 2.72263527e-01, 2.81718552e-01, 1.38080209e-01,
            -1.27307668e-01]
        data = read()[0].data
        template = data[400:600]
        data = data[380:620]
        cc = correlate_template(data, template, method='direct')
        np.testing.assert_allclose(cc, result, atol=1e-7)
        shift, corr = xcorr_max(cc)
        assert round(abs(corr-1.0), 7) == 0
        assert shift == 0

    def test_correlate_template_different_normalizations(self):
        data = read()[0].data
        template = data[400:600]
        data = data[380:700]
        max_index = 20
        ct = correlate_template
        full_xcorr = ct(data, template, demean=False)
        naive_xcorr = ct(data, template, demean=False, normalize='naive')
        nonorm_xcorr = ct(data, template, demean=False, normalize=None)
        assert np.argmax(full_xcorr) == max_index
        assert np.argmax(naive_xcorr) == max_index
        assert np.argmax(nonorm_xcorr) == max_index
        assert round(abs(full_xcorr[max_index]-1.0), 7) == 0
        assert naive_xcorr[max_index] < full_xcorr[max_index]
        np.testing.assert_allclose(nonorm_xcorr, np.correlate(data, template))

    def test_correlate_template_correct_alignment_of_normalization(self):
        data = read()[0].data
        template = data[400:600]
        data = data[380:620]
        # test for all combinations of odd and even length input data
        for i1, i2 in ((0, 0), (0, 1), (1, 1), (1, 0)):
            for mode in ('valid', 'same', 'full'):
                for demean in (True, False):
                    xcorr = correlate_template(data[i1:], template[i2:],
                                               mode=mode, demean=demean)
                    assert round(abs(np.max(xcorr)-1), 7) == 0

    def test_correlate_template_versus_correlate(self):
        data = read()[0].data
        template = data[400:600]
        data = data[380:620]
        xcorr1 = correlate_template(data, template, normalize='naive')
        xcorr2 = correlate(data, template, 20)
        np.testing.assert_equal(xcorr1, xcorr2)

    def test_correlate_template_zeros_in_input(self):
        template = np.zeros(10)
        data = read()[0].data[380:420]
        xcorr = correlate_template(data, template)
        np.testing.assert_equal(xcorr, np.zeros(len(xcorr)))
        template[:] = data[:10]
        data[5:20] = 0
        xcorr = correlate_template(data, template)
        np.testing.assert_equal(xcorr[5:11], np.zeros(6))
        data[:] = 0
        xcorr = correlate_template(data, template)
        np.testing.assert_equal(xcorr, np.zeros(len(xcorr)))
        xcorr = correlate_template(data, template, normalize='naive')
        np.testing.assert_equal(xcorr, np.zeros(len(xcorr)))

    def test_correlate_template_different_amplitudes(self):
        """
        Check that correlations are the same independent of template amplitudes
        """
        data = np.random.randn(20000)
        template = data[1000:1200]
        template_large = template * 10e10
        template_small = template * 10e-10

        cc = correlate_template(data, template)
        cc_large = correlate_template(data, template_large)
        cc_small = correlate_template(data, template_small)
        np.testing.assert_allclose(cc, cc_large)
        np.testing.assert_allclose(cc, cc_small)

    def test_correlate_template_nodemean_fastmatchedfilter(self):
        """
        Compare non-demeaned result against FMF derived result.

        FMF result obtained by the following:

        import copy
        import numpy as np
        from fast_matched_filter import matched_filter
        from obspy import read

        data = read()[0].data
        template = copy.deepcopy(data[400:600])
        data = data[380:620]
        result = matched_filter(
            templates=template.reshape(1, 1, 1, len(template)),
            moveouts=np.array(0).reshape(1, 1, 1),
            weights=np.array(1).reshape(1, 1, 1),
            data=data.reshape(1, 1, len(data)),
            step=1, arch='cpu')[0]

        .. note::
            FastMatchedFilter doesn't use semver, but result generated by Calum
            Chamberlain on 18 Jan 2018 using up-to-date code, with the patch
            in https://github.com/beridel/fast_matched_filter/pull/12
        """
        result = [
            -1.48108244e-01, 4.71532270e-02, 1.82797655e-01,
            1.92574233e-01, 1.18700281e-01, 1.18958903e-02,
            -9.23405439e-02, -1.40047163e-01, -1.00863703e-01,
            -4.86961426e-03, 1.04124829e-01, 1.72662303e-01,
            1.41110823e-01, 1.53776666e-04, -1.71214968e-01,
            -2.83201426e-01, -3.04899812e-01, -2.03215942e-01,
            8.88349637e-02, 5.00749528e-01, 7.18140483e-01,
            5.29728174e-01, 1.30591258e-01, -1.83402568e-01,
            -3.22406143e-01, -3.20676118e-01, -1.98054180e-01,
            -5.06028766e-04, 1.56253457e-01, 1.74580097e-01,
            6.49696961e-02, -8.56237561e-02, -1.89858019e-01,
            -1.96504310e-01, -1.04968190e-01, 2.51029599e-02,
            1.32686019e-01, 2.03692451e-01, 2.11983219e-01,
            0.00000000e+00, 0.00000000e+00]
        data = read()[0].data
        template = data[400:600]
        data = data[380:620]
        # FMF demeans template but does not locally demean data for
        # normalization
        template = template - template.mean()
        cc = correlate_template(data, template, demean=False)
        # FMF misses the last two elements?
        np.testing.assert_allclose(cc[0:-2], result[0:-2], atol=1e-7)
        shift, corr = xcorr_max(cc)
        assert shift == 0

    def test_integer_input_equals_float_input(self):
        a = [-3, 0, 4]
        b = [-3, 4]
        c = np.array(a, dtype=float)
        d = np.array(b, dtype=float)
        for demean in (True, False):
            for normalize in (None, 'naive'):
                cc1 = correlate(a, b, 3, demean=demean, normalize=normalize,
                                method='direct')
                cc2 = correlate(c, d, 3, demean=demean, normalize=normalize)
                np.testing.assert_allclose(cc1, cc2)
            for normalize in (None, 'naive', 'full'):
                cc3 = correlate_template(a, b, demean=demean,
                                         normalize=normalize, method='direct')
                cc4 = correlate_template(c, d, demean=demean,
                                         normalize=normalize)
                np.testing.assert_allclose(cc3, cc4)

    def test_correlate_stream_template_and_correlation_detector(self):
        template = read().filter('highpass', freq=5).normalize()
        pick = UTCDateTime('2009-08-24T00:20:07.73')
        template.trim(pick, pick + 10)
        n1 = len(template[0])
        n2 = 100 * 3600  # 1 hour
        dt = template[0].stats.delta
        # shift one template Trace
        template[1].stats.starttime += 5
        stream = template.copy()
        np.random.seed(42)
        for tr, trt in zip(stream, template):
            tr.stats.starttime += 24 * 3600
            tr.data = np.random.random(n2) - 0.5  # noise
            if tr.stats.channel[-1] == 'Z':
                tr.data[n1:2 * n1] += 10 * trt.data
                tr.data = tr.data[:-n1]
            tr.data[5 * n1:6 * n1] += 100 * trt.data
            tr.data[20 * n1:21 * n1] += 2 * trt.data
        # make one template trace a bit shorter
        template[2].data = template[2].data[:-n1 // 5]
        # make two stream traces a bit shorter
        stream[0].trim(5, None)
        stream[1].trim(1, 20)
        # second template
        pick2 = stream[0].stats.starttime + 20 * n1 * dt
        template2 = stream.slice(pick2 - 5, pick2 + 5)
        # test cross correlation
        stream_orig = stream.copy()
        template_orig = template.copy()
        ccs = correlate_stream_template(stream, template)
        assert len(ccs) == len(stream)
        assert stream[1].stats.starttime == ccs[0].stats.starttime
        assert stream_orig == stream
        assert template_orig == template
        # test if traces with not matching seed ids are discarded
        ccs = correlate_stream_template(stream[:2], template[1:])
        assert len(ccs) == 1
        assert stream_orig == stream
        assert template_orig == template
        # test template_time parameter
        ccs1 = correlate_stream_template(stream, template)
        template_time = template[0].stats.starttime + 100
        ccs2 = correlate_stream_template(stream, template,
                                         template_time=template_time)
        assert len(ccs2) == len(ccs1)
        delta = ccs2[0].stats.starttime - ccs1[0].stats.starttime
        assert round(abs(delta-100), 7) == 0
        # test if all three events found
        detections, sims = correlation_detector(stream, template, 0.2, 30)
        assert len(detections) == 3
        dtime = pick + n1 * dt + 24 * 3600
        assert round(abs(detections[0]['time']-dtime), 7) == 0
        assert len(sims) == 1
        assert stream_orig == stream
        assert template_orig == template
        # test if xcorr stream is suitable for coincidence_trigger
        # result should be the same, return values related
        ccs = correlate_stream_template(stream, template)
        triggers = coincidence_trigger(None, 0.2, -1, ccs, 2,
                                       max_trigger_length=30, details=True)
        assert len(triggers) == 2
        for d, t in zip(detections[1:], triggers):
            assert round(abs(np.mean(t['cft_peaks'])-d['similarity']), 7) == 0
        # test template_magnitudes
        detections, _ = correlation_detector(stream, template, 0.2, 30,
                                             template_magnitudes=1)
        assert abs(detections[1]['amplitude_ratio']-100) < 1
        assert abs(detections[1]['magnitude'] - (1 + 8 / 3)) < 0.01
        assert abs(detections[2]['amplitude_ratio']-2) < 2
        detections, _ = correlation_detector(stream, template, 0.2, 30,
                                             template_magnitudes=True)
        assert abs(detections[1]['amplitude_ratio']-100) < 1
        assert 'magnitude' not in detections[1]
        assert stream_orig == stream
        assert template_orig == template
        # test template names
        detections, _ = correlation_detector(stream, template, 0.2, 30,
                                             template_names='eq')
        assert detections[0]['template_name'] == 'eq'
        detections, _ = correlation_detector(stream, template, 0.2, 30,
                                             template_names=['eq'], plot=True)
        assert detections[0]['template_name'] == 'eq'
        # test similarity parameter with additional constraints
        # test details=True

        def simf(ccs):
            ccmatrix = np.array([tr.data for tr in ccs])
            comp_thres = np.sum(ccmatrix > 0.2, axis=0) > 1
            similarity = ccs[0].copy()
            similarity.data = np.mean(ccmatrix, axis=0) * comp_thres
            return similarity
        detections, _ = correlation_detector(stream, template, 0.1, 30,
                                             similarity_func=simf,
                                             details=True)
        assert len(detections) == 2
        for d in detections:
            mean_val = np.mean(list(d['cc_values'].values()))
            assert round(abs(mean_val - d['similarity']), 7) == 0
        # test if properties from find_peaks function are returned
        detections, sims = correlation_detector(stream, template, 0.1, 30,
                                                threshold=0.16, details=True,
                                                similarity_func=simf)
        try:
            from scipy.signal import find_peaks  # noqa
        except ImportError:
            assert len(detections) == 2
            assert 'left_threshold' not in detections[0]
        else:
            assert len(detections) == 1
            assert 'left_threshold' in detections[0]
        # also check the _find_peaks function
        distance = int(round(30 / sims[0].stats.delta))
        indices = _find_peaks(sims[0].data, 0.1, distance, distance)
        assert len(indices) == 2
        # test distance parameter
        detections, _ = correlation_detector(stream, template, 0.2, 500)
        assert len(detections) == 1
        # test more than one template
        # just 2 detections for first template, because second template has
        # a higher similarity for third detection
        templates = (template, template2)
        templatetime2 = pick2 - 10
        template_times = (template[0].stats.starttime, templatetime2)
        detections, _ = correlation_detector(stream, templates, (0.2, 0.3), 30,
                                             plot=stream,
                                             template_times=template_times,
                                             template_magnitudes=(2, 5))
        assert len(detections) > 0
        assert 'template_id' in detections[0]
        detections0 = [d for d in detections if d['template_id'] == 0]
        assert len(detections0) == 2
        assert len(detections) == 3
        assert round(abs(detections[2]['similarity']-1), 7) == 0
        assert round(abs(detections[2]['magnitude']-5), 7) == 0
        assert detections[2]['time'] == templatetime2
        # test if everything is correct if template2 and stream do not have
        # any ids in common
        templates = (template, template2[2:])
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            detections, sims = correlation_detector(
                stream[:1], templates, 0.2, 30, plot=True,
                template_times=templatetime2, template_magnitudes=2)
        detections0 = [d for d in detections if d['template_id'] == 0]
        assert len(detections0) == 3
        assert len(detections) == 3
        assert len(sims) == 2
        assert isinstance(sims[0], Trace)
        assert sims[1] is None



# -*- coding: utf-8 -*-
"""
GSE2/GSE1 bindings to ObsPy core module.
"""
import numpy as np

from obspy import Stream, Trace
from . import libgse1, libgse2
from obspy.core.util import open_bytes_stream


def _is_gse2(file):
    """
    Checks whether a file is GSE2 or not.

    :type file: str or file-like object
    :param file: GSE2 file to be checked.
    :rtype: bool
    :return: ``True`` if a GSE2 file.
    """
    with open_bytes_stream(file) as io_stream:
        try:
            libgse2.is_gse2(io_stream)
        except Exception:
            return False
    return True


def _read_gse2(file, headonly=False, verify_chksum=True,
               **kwargs):  # @UnusedVariable
    """
    Reads a GSE2 file and returns a Stream object.

    GSE2 files containing multiple WID2 entries/traces are supported.

    .. warning::
        This function should NOT be called directly, it registers via the
        ObsPy :func:`~obspy.core.stream.read` function, call this instead.

    :type file: str or file-like object
    :param file: GSE2 file to be read.
    :type headonly: bool, optional
    :param headonly: If True read only head of GSE2 file.
    :type verify_chksum: bool, optional
    :param verify_chksum: If True verify Checksum and raise Exception if
        it is not correct.
    :rtype: :class:`~obspy.core.stream.Stream`
    :returns: Stream object containing header and data.

    .. rubric:: Example

    >>> from obspy import read
    >>> st = read("/path/to/loc_RJOB20050831023349.z")
    """
    traces = []
    with open_bytes_stream(file) as f:
        # reading multiple gse2 parts
        while True:
            try:
                if headonly:
                    header = libgse2.read_header(f)
                    traces.append(Trace(header=header))
                else:
                    header, data = libgse2.read(f, verify_chksum=verify_chksum)
                    traces.append(Trace(header=header, data=data))
            except EOFError:
                break
    return Stream(traces=traces)


def _write_gse2(stream, filename, inplace=False, **kwargs):  # @UnusedVariable
    """
    Write GSE2 file from a Stream object.

    .. warning::
        This function should NOT be called directly, it registers via the
        the :meth:`~obspy.core.stream.Stream.write` method of an
        ObsPy :class:`~obspy.core.stream.Stream` object, call this instead.

    :type stream: :class:`~obspy.core.stream.Stream`
    :param stream: The ObsPy Stream object to write.
    :type filename: str
    :param filename: Name of file to write.
    :type inplace: bool, optional
    :param inplace: If True, do compression not on a copy of the data but
        on the data itself - note this will change the data values and make
        them therefore unusable!

    .. rubric:: Example

    >>> from obspy import read
    >>> st = read()
    >>> st.write('filename.gse', format='GSE2') #doctest: +SKIP
    """
    #
    # Translate the common (renamed) entries
    with open(filename, 'wb') as f:
        # write multiple gse2 parts
        for trace in stream:
            dt = np.dtype(np.int32)
            if trace.data.dtype.name == dt.name:
                trace.data = np.ascontiguousarray(trace.data, dt)
            else:
                msg = "GSE2 data must be of type %s, but are of type %s" % \
                    (dt.name, trace.data.dtype)
                raise Exception(msg)
            libgse2.write(trace.stats, trace.data, f, inplace)


def _is_gse1(file):
    """
    Checks whether a file is GSE1 or not.

    :type file: str or file-like object
    :param file: GSE1 file to be checked.
    :rtype: bool
    :return: ``True`` if a GSE1 file.
    """
    with open_bytes_stream(file) as io_stream:
        try:
            data = io_stream.readline()
        except Exception:
            return False
    return data.startswith(b'WID1') or data.startswith(b'XW01')


def _read_gse1(file, headonly=False, verify_chksum=True,
               **kwargs):  # @UnusedVariable
    """
    Reads a GSE1 file and returns a Stream object.

    GSE1 files containing multiple WID1 entries/traces are supported.

    .. warning::
        This function should NOT be called directly, it registers via the
        ObsPy :func:`~obspy.core.stream.read` function, call this instead.

    :type file: str or file-like object
    :param file: GSE2 file to be read.
    :type headonly: bool, optional
    :param headonly: If True read only header of GSE1 file.
    :type verify_chksum: bool, optional
    :param verify_chksum: If True verify Checksum and raise Exception if
        it is not correct.
    :rtype: :class:`~obspy.core.stream.Stream`
    :returns: Stream object containing header and data.

    .. rubric:: Example

    >>> from obspy import read
    >>> st = read("/path/to/y2000.gse")
    """
    traces = []
    # read GSE1 file
    with open_bytes_stream(file) as fh:
        while True:
            try:
                if headonly:
                    header = libgse1.read_header(fh)
                    traces.append(Trace(header=header))
                else:
                    header, data = \
                        libgse1.read(fh, verify_chksum=verify_chksum)
                    traces.append(Trace(header=header, data=data))
            except EOFError:
                break
    return Stream(traces=traces)



# -*- coding: utf-8 -*-
import pytest

from obspy import read_events, Catalog, UTCDateTime
from obspy.core.event import Event
from obspy.io.focmec.core import _is_focmec, _read_focmec


lst_file_first_comment = '\n'.join((
    "",
    "     Dip,Strike,Rake     76.43    59.08   -64.23",
    "     Dip,Strike,Rake     28.90   174.99  -150.97   Auxiliary Plane",
    "     Lower Hem. Trend, Plunge of A,N     84.99    61.10   329.08    "
    "13.57",
    "     Lower Hem. Trend, Plunge of P,T    358.82    51.71   128.90    "
    "26.95",
    "     B trend, B plunge, Angle:  232.62  25.00 105.00",
    "",
    "          Log10(Ratio)                              Ratio     S Polarity",
    "     Observed  Calculated    Difference  Station     Type     Obs.  "
    "Calc. Flag",
    "      0.8847      0.8950      -0.0103      BLA        SH       R      "
    "R       ",
    "      1.1785      1.0810       0.0975      COR        SH       R      "
    "R       ",
    "      0.6013      0.5442       0.0571      HRV        SH       R      "
    "R       ",
    "      0.3287      0.3666      -0.0379      KEV        SH       L      "
    "L       ",
    "      0.8291      0.9341      -0.1050      KIP        SH       R      "
    "R       ",
    "      0.8033      0.7815       0.0218      KIP        SV       B      "
    "B       ",
    "      1.0783      1.1857      -0.1074      PAS        SH       R      "
    "R       ",
    "      0.2576      0.2271       0.0305      TOL        SH       L      "
    "L       ",
    "     -0.2762     -0.4076       0.1314      TOL        SS       F      "
    "F    NUM",
    "     -0.4283     -0.4503       0.0220      HRV        SS       F      "
    "F       ",
    "     -0.0830      0.0713      -0.1543      KEV        SS       B      "
    "B       ",
    "",
    "Total number of ratios used is  11",
    "RMS error for the 11 acceptable solutions is 0.0852",
    "  Highest absolute velue of diff for those solutions is  0.1543"))
out_file_first_comment = (
    "    Dip   Strike   Rake    Pol: P     SV    SH  AccR/TotR  RMS RErr  "
    "AbsMaxDiff\n   76.43   59.08  -64.23       0.00  0.00  0.00    11/11    "
    "0.0852    0.1543")
creation_time = UTCDateTime(2017, 9, 8, 14, 54, 58)


class TestFOCMEC():
    """
    Test everything related to reading FOCMEC files
    """
    @pytest.fixture(autouse=True, scope="function")
    def setup(self, testdata):
        self.lst_file = testdata['focmec_8sta.lst']
        self.out_file = testdata['focmec_8sta.out']
        with open(self.out_file, 'rb') as fh:
            header = []
            for i in range(15):
                header.append(fh.readline().decode('ASCII'))
            self.out_file_header = ''.join(header).rstrip()
        with open(self.lst_file, 'rb') as fh:
            header = []
            for i in range(48):
                header.append(fh.readline().decode('ASCII'))
            self.lst_file_header = ''.join(header).rstrip()

    def _assert_cat_common_parts(self, cat):
        assert isinstance(cat, Catalog)
        assert len(cat) == 1
        event = cat[0]
        assert isinstance(event, Event)
        assert event.creation_info.creation_time == \
            UTCDateTime(2017, 9, 8, 14, 54, 58)
        assert len(event.focal_mechanisms) == 4
        expected_dip_strike_rake = (
            (76.43, 59.08, -64.23),
            (77.05, 54.08, -59.13),
            (77.05, 59.89, -59.13),
            (77.76, 54.50, -54.06))
        for focmec, (dip, strike, rake) in zip(
                event.focal_mechanisms, expected_dip_strike_rake):
            plane1 = focmec.nodal_planes.nodal_plane_1
            assert plane1.strike == strike
            assert plane1.dip == dip
            assert plane1.rake == rake
            assert focmec.nodal_planes.preferred_plane == 1
        for focmec in cat[0].focal_mechanisms:
            assert focmec.station_polarity_count == 23
            # check creation time
            assert focmec.creation_info.creation_time == creation_time
            # check creation info version
            assert focmec.creation_info.version == 'FOCMEC'

    def _assert_cat_out(self, cat):
        self._assert_cat_common_parts(cat)
        assert cat[0].comments[0].text == self.out_file_header
        assert cat[0].focal_mechanisms[0].comments[0].text == \
            out_file_first_comment
        for focmec in cat[0].focal_mechanisms:
            # misfit should be None, because the file specifies that polarity
            # errors are weighted and in the out file format we can't know how
            # many individual errors there are
            assert focmec.misfit is None
            # we can't tell the gap from the out format
            assert focmec.azimuthal_gap is None

    def _assert_cat_lst(self, cat):
        self._assert_cat_common_parts(cat)
        assert cat[0].comments[0].text == self.lst_file_header
        assert cat[0].focal_mechanisms[0].comments[0].text == \
            lst_file_first_comment
        for focmec in cat[0].focal_mechanisms:
            # misfit should be 0.0, because in the lst file we can count the
            # number of individual errors (and there's no polarity errors)
            assert focmec.misfit == 0.0
            # we can't tell the gap from the out format
            assert focmec.azimuthal_gap == 236.7

    def test_is_focmec(self):
        for file_ in (self.lst_file, self.out_file):
            assert _is_focmec(file_)

    def test_read_focmec_out(self):
        cat = _read_focmec(self.out_file)
        self._assert_cat_out(cat)

    def test_read_focmec_out_open_file(self):
        for mode in ('rb', 'rt'):
            with open(self.out_file, mode) as fh:
                cat = _read_focmec(fh)
            self._assert_cat_out(cat)

    def test_read_focmec_lst(self):
        cat = _read_focmec(self.lst_file)
        self._assert_cat_lst(cat)

    def test_read_focmec_lst_open_file(self):
        for mode in ('rb', 'rt'):
            with open(self.lst_file, mode) as fh:
                cat = _read_focmec(fh)
            self._assert_cat_lst(cat)

    def test_read_focmec_out_through_plugin(self):
        cat = read_events(self.out_file)
        self._assert_cat_out(cat)

    def test_read_focmec_lst_through_plugin(self):
        cat = read_events(self.lst_file)
        self._assert_cat_lst(cat)

    def test_read_focmec_lst_other_flavors(self, testdata):
        """
        This tests some additional files. lst files have a pretty free format
        unfortunately. It depends on focmec program version and on input data /
        program options what lines get printed and also how specific lines we
        look for look like..

        For these files just test a few things, in general reading should be
        covered by basic tests above
        """
        # 1: focmec_qedUWne.lst
        cat = read_events(testdata['focmec_qedUWne.lst'])
        assert len(cat) == 1
        focmecs = cat[0].focal_mechanisms
        assert len(focmecs) == 5
        focmec = focmecs[0]
        plane = focmec.nodal_planes.nodal_plane_1
        assert plane.strike == 308.43
        assert plane.dip == 58.68
        assert plane.rake == 16.48
        for focmec in focmecs:
            assert focmec.azimuthal_gap == 79.5
        for focmec in focmecs[:-1]:
            assert focmec.station_polarity_count == 190
            assert focmec.misfit == 21.0 / 190
        # last solution has one less polarity error
        assert focmecs[-1].station_polarity_count == 190
        assert focmecs[-1].misfit == 20.0 / 190

        # 2: focmec_all.lst
        cat = read_events(testdata['focmec_all.lst'])
        assert len(cat) == 1
        focmecs = cat[0].focal_mechanisms
        assert len(focmecs) == 3
        focmec = focmecs[0]
        plane = focmec.nodal_planes.nodal_plane_1
        assert plane.strike == 66.98
        assert plane.dip == 80.61
        assert plane.rake == -69.72
        for focmec, errors in zip(focmecs, (25, 29, 26)):
            assert focmec.azimuthal_gap == 44.7
            assert focmec.station_polarity_count == 212
            assert focmec.misfit == float(errors) / 212

        # 3: focmec_8sta-noratios.lst
        cat = read_events(testdata['focmec_8sta-noratios.lst'])
        assert len(cat) == 1
        focmecs = cat[0].focal_mechanisms
        assert len(focmecs) == 184
        focmec = focmecs[2]
        plane = focmec.nodal_planes.nodal_plane_1
        assert plane.strike == 255.00
        assert plane.dip == 10.0
        assert plane.rake == -90.00
        for focmec in focmecs:
            assert focmec.azimuthal_gap == 236.7
            assert focmec.station_polarity_count == 23
            assert focmec.misfit == 0.0



# -*- coding: utf-8 -*-
"""
The obspy.imaging.source test suite.
"""
from obspy.imaging.source import plot_radiation_pattern


class TestRadPattern:
    """
    Test cases for radiation_pattern.
    """
    def test_farfield_with_quiver(self):
        """
        Tests to plot P/S wave farfield radiation pattern
        """
        # Peru 2001/6/23 20:34:23:
        mt = [2.245, -0.547, -1.698, 1.339, -3.728, 1.444]
        plot_radiation_pattern(
            mt, kind=['beachball', 's_quiver', 'p_quiver'], show=False)



# -*- coding: utf-8 -*-
"""
Tests for reading rg16 format.
"""
import io

import numpy as np
import pytest

import obspy
import obspy.io.rg16.core as rc
from obspy import read, UTCDateTime


class TestReadRG16():

    @pytest.fixture(autouse=True, scope="function")
    def setup(self, testdata):
        self.one_chan_fcnt = testdata['one_channel_many_traces.fcnt']
        self.three_chan_fcnt = testdata['three_chans_six_traces.fcnt']
        self.bad_alias_filter = testdata['channel_set_bad_alias_filter.dat']
        self.header_block_same_ru_code = testdata['header_3_chan_one_code.dat']
        self.fcnt_files = [self.one_chan_fcnt, self.three_chan_fcnt]

    def test_reading_rg16_files(self):
        """
        Ensure that the rg16 files are read by the function
        :func:`~obspy.core.stream.read` with or without specifying the format.
        """
        for fcnt_file in self.fcnt_files:
            st_1 = read(fcnt_file)
            st_2 = read(fcnt_file, format="RG16")
            st_3 = rc._read_rg16(fcnt_file)
            # when the function read is called a key "_format" is introduced
            # in the object stats. This key is not created when the function
            # _read_rg16 is called. In order to check the stream equality, the
            # key "_format" was removed.
            for tr_1, tr_2 in zip(st_1, st_2):
                del tr_1.stats._format
                del tr_2.stats._format
            assert st_1 == st_3
            assert st_2 == st_3

    def test_rg16_files_identified(self):
        """
        Ensure the rg16 files are correctly labeled as such.
        """
        for fcnt_file in self.fcnt_files:
            assert rc._is_rg16(fcnt_file)
            with open(fcnt_file, 'rb') as fi:
                assert rc._is_rg16(fi)

    def test_empty_buffer(self):
        """
        Ensure an empty buffer returns false.
        """
        buff = io.BytesIO()
        assert not rc._is_rg16(buff)

    def test_headonly_option(self):
        """
        Ensure no data is returned when the option headonly is used.
        """
        st = rc._read_rg16(self.three_chan_fcnt, headonly=True)
        for tr in st:
            assert len(tr.data) == 0
            assert tr.stats.npts != 0

    def test_starttime_endtime_option(self):
        """
        Test the options starttime and endtime
        """
        t1 = UTCDateTime(2017, 8, 9, 16, 0, 15)
        t2 = UTCDateTime(2017, 8, 9, 16, 0, 45)

        # read streams for testing. The three channel rg16 file has 6 traces
        # but the streams may have less depending on the starttime/endtime
        st = rc._read_rg16(self.three_chan_fcnt)  # no time filtering
        st1 = rc._read_rg16(self.three_chan_fcnt, starttime=t1, endtime=t2)

        # test using starttime and endtime
        assert len(st1) == len(st)
        for tr, tr1 in zip(st, st1):
            assert tr.stats.starttime == tr1.stats.starttime
            assert tr.stats.endtime == tr1.stats.endtime

    def test_intrablock_starttime_endtime(self):
        """
        Test starttime/endtime options when starttime and endtime are comprised
        in a data block.
        """
        t1 = UTCDateTime(2017, 8, 9, 16, 0, 47)
        t2 = UTCDateTime(2017, 8, 9, 16, 0, 58)

        # read streams for testing
        st = rc._read_rg16(self.three_chan_fcnt)  # no time filtering
        st1 = rc._read_rg16(self.three_chan_fcnt, starttime=t1, endtime=t2)

        # test when starttime and endtime are comprised in a data packet.
        assert len(st1) == 3
        for tr, tr1 in zip(st[1::2], st1):
            assert tr.stats.starttime == tr1.stats.starttime
            assert tr.stats.endtime == tr1.stats.endtime

    def test_merge(self):
        """
        Ensure the merge option of read_rg16 merges all contiguous traces
        together.
        """
        for fcnt_file in self.fcnt_files:
            st_merged = rc._read_rg16(fcnt_file, merge=True)
            st = rc._read_rg16(fcnt_file).merge()
            assert len(st) == len(st_merged)
            assert st == st_merged

    def test_contacts_north_and_merge(self):
        """
        Ensure the "contacts_north" and "merge" parameters can be used
        together. See #2198.
        """
        for filename in self.fcnt_files:
            st = rc._read_rg16(filename, contacts_north=True, merge=True)
            assert isinstance(st, obspy.Stream)

    def test_make_stats(self):
        """
        Check function make_stats.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            stats = rc._make_stats(fi, 288, False, False)
        assert isinstance(stats, obspy.core.trace.Stats)
        assert abs(stats.sampling_rate-500) < 1e-5
        assert stats.network == '1'
        assert stats.station == '1'
        assert stats.location == '1'
        assert stats.channel == 'DP3'
        assert stats.npts == 15000
        assert stats.starttime == UTCDateTime('2017-08-09T16:00:00.380000Z')
        assert stats.endtime == UTCDateTime('2017-08-09T16:00:30.378000Z')
        with open(self.three_chan_fcnt, 'rb') as fi:
            stats = rc._make_stats(fi, 288, True, False)
        assert stats.channel == 'DPN'


class TestReadRG16Headers():

    @pytest.fixture(autouse=True, scope="function")
    def setup(self, testdata):
        self.one_chan_fcnt = testdata['one_channel_many_traces.fcnt']
        self.three_chan_fcnt = testdata['three_chans_six_traces.fcnt']
        self.bad_alias_filter = testdata['channel_set_bad_alias_filter.dat']
        self.header_block_same_ru_code = testdata['header_3_chan_one_code.dat']
        self.fcnt_files = [self.one_chan_fcnt, self.three_chan_fcnt]

    def test_cmp_nbr_headers(self):
        """
        Test to check that the number of headers is correct.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            (nbr_channel_set_headers, nbr_extended_headers,
             nbr_external_headers) = rc._cmp_nbr_headers(fi)
        assert nbr_channel_set_headers == 3
        assert nbr_extended_headers == 3
        assert nbr_external_headers == 1

    def test_cmp_nbr_records(self):
        """
        Check the number of records in the file.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            nbr_records = rc._cmp_nbr_records(fi)
        assert nbr_records == 6
        with open(self.one_chan_fcnt, 'rb') as fi:
            nbr_records = rc._cmp_nbr_records(fi)
        assert nbr_records == 10

    def test_cmp_jump(self):
        """
        Check the number of bytes to jump to reach the next trace block.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            jump = rc._cmp_jump(fi, 288)
        assert jump == 60340

    def test_read_trace_header(self):
        """
        Test the reading of the trace header.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            trace_header = rc._read_trace_header(fi, 288)
        assert trace_header['trace_number'] == 1
        assert trace_header['trace_edit_code'] == 0

    def test_read_trace_header_1(self):
        """
        Test the reading of the trace header 1.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            trace_header_1 = rc._read_trace_header_1(fi, 288)
        assert trace_header_1['extended_receiver_line_nbr'] == 65536
        assert trace_header_1['extended_receiver_point_nbr'] == 65536
        assert trace_header_1['sensor_type'] == 3
        assert trace_header_1['trace_count_file'] == 1

    def test_read_trace_header_2(self):
        """
        Test the reading of the trace header 2.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            trace_header_2 = rc._read_trace_header_2(fi, 288)
        assert trace_header_2['shot_line_nbr'] == 2240
        assert trace_header_2['shot_point'] == 1
        assert trace_header_2['shot_point_index'] == 0
        assert abs(trace_header_2['shot_point_pre_plan_x']-0) < 1e-5
        assert abs(trace_header_2['shot_point_pre_plan_y']-0) < 1e-5
        assert abs(trace_header_2['shot_point_final_x']-0) < 1e-5
        assert abs(trace_header_2['shot_point_final_y']-0) < 1e-5
        assert abs(trace_header_2['shot_point_final_depth']-0) < 1e-5
        assert trace_header_2['source_of_final_shot_info'] == 'undefined'
        assert trace_header_2['energy_source_type'] == 'undefined'

    def test_read_trace_header_3(self):
        """
        Test the reading of the trace header 3.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            trace_header_3 = rc._read_trace_header_3(fi, 288)
        assert trace_header_3['epoch_time'] == \
            UTCDateTime('2017-08-09T16:00:00.380000Z')
        assert abs(trace_header_3['shot_skew_time']-0) < 1e-5
        assert abs(trace_header_3['time_shift_clock_correction']-0) < 1e-5
        assert abs(trace_header_3['remaining_clock_correction']-0) < 1e-5

    def test_read_trace_header_4(self):
        """
        Test the reading of the trace header 4.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            trace_header_4 = rc._read_trace_header_4(fi, 288)
        assert abs(trace_header_4['pre_shot_guard_band']-0) < 1e-5
        assert abs(trace_header_4['post_shot_guard_band']-0) < 1e-5
        assert trace_header_4['preamp_gain'] == 24
        assert trace_header_4['trace_clipped_flag'] == 'not clipped'
        assert trace_header_4['record_type_code'] == \
            'normal seismic data record'
        assert trace_header_4['shot_status_flag'] == 'normal'
        assert trace_header_4['external_shot_id'] == 0
        key = 'post_processed_first_break_pick_time'
        assert abs(trace_header_4[key]-0) < 1e-5
        assert abs(trace_header_4['post_processed_rms_noise']-0) < 1e-5

    def test_read_trace_header_5(self):
        """
        Test the reading of the trace header 5.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            trace_header_5 = rc._read_trace_header_5(fi, 288)
        assert abs(trace_header_5['receiver_point_pre_plan_x']-469567.2) < 1e-5
        assert \
            abs(trace_header_5['receiver_point_pre_plan_y']-5280707.8) < 1e-5
        assert abs(trace_header_5['receiver_point_final_x']-469565.2) < 1e-5
        assert abs(trace_header_5['receiver_point_final_y']-5280709.7) < 1e-5
        assert abs(trace_header_5['receiver_point_final_depth']-0) < 1e-5
        assert trace_header_5['source_of_final_receiver_info'] == \
            'as laid (no navigation sensor)'

    def test_read_trace_header_6(self):
        """
        Test the reading of the trace header 6.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            trace_header_6 = rc._read_trace_header_6(fi, 288)
        assert abs(trace_header_6['tilt_matrix_h1x']-0) < 1e-5
        assert abs(trace_header_6['tilt_matrix_h2x']-0) < 1e-5
        assert abs(trace_header_6['tilt_matrix_vx']-0) < 1e-5
        assert abs(trace_header_6['tilt_matrix_h1y']-0) < 1e-5
        assert abs(trace_header_6['tilt_matrix_h2y']-0) < 1e-5
        assert abs(trace_header_6['tilt_matrix_vy']-0) < 1e-5
        assert abs(trace_header_6['tilt_matrix_h1z']-0) < 1e-5
        assert abs(trace_header_6['tilt_matrix_h2z']-0) < 1e-5

    def test_read_trace_header_7(self):
        """
        Test the reading of the trace header 7.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            trace_header_7 = rc._read_trace_header_7(fi, 288)
        assert abs(trace_header_7['tilt_matrix_vz']-0) < 1e-5
        assert abs(trace_header_7['azimuth_degree']-0) < 1e-5
        assert abs(trace_header_7['pitch_degree']-0) < 1e-5
        assert abs(trace_header_7['roll_degree']-0) < 1e-5
        assert abs(trace_header_7['remote_unit_temp']-0) < 1e-5
        assert abs(trace_header_7['remote_unit_humidity']-0) < 1e-5
        assert trace_header_7['orientation_matrix_version_nbr'] == 0
        assert trace_header_7['gimbal_corrections'] == 0

    def test_read_trace_header_8(self):
        """
        Test the reading of the trace header 8.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            trace_header_8 = rc._read_trace_header_8(fi, 288)
        assert trace_header_8['fairfield_test_analysis_code'] == 0
        assert trace_header_8['first_test_oscillator_attenuation'] == 0
        assert trace_header_8['second_test_oscillator_attenuation'] == 0
        assert abs(trace_header_8['start_delay']-0) < 1e-5
        assert trace_header_8['dc_filter_flag'] == 0
        assert abs(trace_header_8['dc_filter_frequency']-0) < 1e-5
        assert trace_header_8['preamp_path'] == 'external input selected'
        assert trace_header_8['test_oscillator_signal_type'] == \
            'test oscillator path open'

    def test_read_trace_header_9(self):
        """
        Test the reading of the trace header 9.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            trace_header_9 = rc._read_trace_header_9(fi, 288)
        assert trace_header_9['test_signal_generator_signal_type'] == \
            'pattern is address ramp'
        key = 'test_signal_generator_frequency_1'
        assert abs(trace_header_9[key]-0) < 1e-5
        key = 'test_signal_generator_frequency_2'
        assert abs(trace_header_9[key]-0) < 1e-5
        assert trace_header_9['test_signal_generator_amplitude_1'] == 0
        assert trace_header_9['test_signal_generator_amplitude_2'] == 0
        key = 'test_signal_generator_duty_cycle_percentage'
        assert abs(trace_header_9[key]-0) < 1e-5
        key = 'test_signal_generator_active_duration'
        assert abs(trace_header_9[key]-0) < 1e-5
        key = 'test_signal_generator_activation_time'
        assert abs(trace_header_9[key]-0) < 1e-5

    def test_read_trace_header_10(self):
        """
        Test the reading of the trace header 10.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            trace_header_10 = rc._read_trace_header_10(fi, 288)
        assert trace_header_10['test_signal_generator_idle_level'] == 0
        assert trace_header_10['test_signal_generator_active_level'] == 0
        assert trace_header_10['test_signal_generator_pattern_1'] == 0
        assert trace_header_10['test_signal_generator_pattern_2'] == 0

    def test_make_trace(self):
        """
        Test if the ten first samples of the waveform are read correctly
        and if the output is a Trace object.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            trace_3 = rc._make_trace(fi, 288, False, False, False)
            trace_z = rc._make_trace(fi, 241648, False, True, False)
        expected = np.array([-0.18864873, -0.30852857, -0.35189095,
                             -0.22547323, -0.12023376, -0.14336781,
                             -0.11712314, 0.04060567, 0.18024819,
                             0.17769636])
        all_close = np.allclose(trace_3.data[:len(expected)], expected)
        assert all_close
        assert isinstance(trace_3, obspy.core.trace.Trace)
        expected = np.array([-0.673309, -0.71590775, -0.54966664, -0.33980238,
                             -0.29999766, -0.3031269, -0.12762846, 0.08782373,
                             0.11377038, 0.09888785])
        all_close = np.allclose(trace_z.data[:len(expected)], expected)
        assert all_close

    def test_can_write(self):
        """
        Ensure the result of _read_rg16 is a stream object and that
        it can be written as mseed.
        """
        st = rc._read_rg16(self.three_chan_fcnt)
        assert isinstance(st, obspy.core.stream.Stream)
        bytstr = io.BytesIO()
        # test passes if this doesn't raise
        try:
            st.write(bytstr, 'mseed')
        except Exception:
            self.fail('Failed to write to mseed!')

    def test_read_general_header_1(self):
        """
        Test the reading of the general header 1.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            header_1 = rc._read_general_header_1(fi)
        assert header_1['base_scan_interval'] == 32
        assert header_1['file_number'] == 1
        assert header_1['general_constant'] == 0
        assert header_1['julian_day'] == 221
        assert header_1['manufacturer_code'] == 20
        assert header_1['manufacturer_serial_number'] == 0
        assert header_1['nbr_add_general_header'] == 1
        assert header_1['nbr_channel_set'] == 3
        assert header_1['nbr_skew_block'] == 0
        assert header_1['polarity_code'] == 0
        assert header_1['record_type'] == 0
        assert header_1['sample_format_code'] == 8058
        assert header_1['scan_type_per_record'] == 1
        assert header_1['time_slice'] == 160000
        assert header_1['time_slice_year'] == 17

    def test_read_general_header_2(self):
        """
        Test the reading of the general header 2.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            header_2 = rc._read_general_header_2(fi)
        assert header_2['extended_channel_sets_per_scan_type'] == 3
        assert header_2['extended_file_number'] == 1
        assert header_2['extended_header_blocks'] == 3
        assert header_2['extended_record_length'] == 30
        assert header_2['external_header_blocks'] == 1
        assert header_2['general_header_block_number'] == 2
        assert header_2['version_number'] == 262

    def test_read_channel_sets(self):
        """
        Test that all the channel sets are read.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            channel_sets = rc._read_channel_sets(fi)
        assert len(channel_sets) == 3

    def test_read_channel_set(self):
        """
        Test the reading of the first channel set.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            channel_set_1 = rc._read_channel_set(fi, 64)
        assert channel_set_1['RU_channel_number'] == 1
        assert channel_set_1['alias_filter_frequency'] == 207
        assert channel_set_1['alias_filter_slope'] == 320
        assert channel_set_1['array_forming'] == 0
        assert abs(channel_set_1['channel_set_end_time']-30) < 1e-5
        assert channel_set_1['channel_set_number'] == 1
        assert abs(channel_set_1['channel_set_start_time']-0) < 1e-5
        assert channel_set_1['channel_type_code'] == 1
        assert channel_set_1['extended_channel_set_number'] == 1
        assert channel_set_1['extended_header_flag'] == 0
        assert channel_set_1['gain_control_type'] == 3
        assert channel_set_1['low_cut_filter_freq'] == 0
        assert channel_set_1['low_cut_filter_slope'] == 6
        assert channel_set_1['mp_factor_descaler_multiplier'] == 0
        assert channel_set_1['nbr_32_byte_trace_header_extension'] == 10
        assert channel_set_1['nbr_channels_in_channel_set'] == 2
        assert abs(channel_set_1['notch_2_filter_freq']-0) < 1e-5
        assert abs(channel_set_1['notch_3_filter_freq']-0) < 1e-5
        assert abs(channel_set_1['notch_filter_freq']-0) < 1e-5
        assert channel_set_1['optionnal_MP_factor'] == 0
        assert channel_set_1['scan_type_number'] == 1
        assert channel_set_1['vertical_stack_size'] == 1

    def test_read_extended_header_1(self):
        """
        Test the reading of the extended header 1.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            extended_header_1 = rc._read_extended_header_1(fi, 160)
        assert extended_header_1['deployment_time'] == \
            UTCDateTime('2017-08-09T15:46:32.230000Z')
        assert extended_header_1['id_ru'] == 1219770716358969536
        assert extended_header_1['pick_up_time'] == \
            UTCDateTime('2017-08-09T20:06:58.120000Z')
        assert extended_header_1['start_time_ru'] == \
            UTCDateTime('2017-08-09T15:52:31.366000Z')

    def test_read_extended_header_2(self):
        """
        Test the reading of the extended header 2.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            extended_header_2 = rc._read_extended_header_2(fi, 192)
        assert abs(extended_header_2['acquisition_drift_window']-0) < 1e-5
        assert abs(extended_header_2['clock_drift']-0) < 1e-5
        assert extended_header_2['clock_stop_method'] == 'normal'
        assert extended_header_2['data_collection_method'] == 'continuous'
        assert extended_header_2['data_decimation'] == 'not decimated'
        assert extended_header_2['file_number'] == 1
        assert extended_header_2['frequency_drift'] == 'within specification'
        assert extended_header_2['nbr_files'] == 1
        assert extended_header_2['nbr_time_slices'] == 2
        key = 'nbr_decimation_filter_coef'
        assert extended_header_2[key] == 0
        assert extended_header_2['original_base_scan_interval'] == 0
        assert extended_header_2['oscillator_type'] == 'disciplined'

    def test_read_extended_header_3(self):
        """
        Test the reading of the extended header 3.
        """
        with open(self.three_chan_fcnt, 'rb') as fi:
            extended_header_3 = rc._read_extended_header_3(fi, 224)
        assert extended_header_3['first_shot_line'] == 0
        assert extended_header_3['first_shot_point'] == 0
        assert extended_header_3['first_shot_point_index'] == 0
        assert extended_header_3['last_shot_line'] == 0
        assert extended_header_3['last_shot_point'] == 0
        assert extended_header_3['last_shot_point_index'] == 0
        assert extended_header_3['receiver_line_number'] == 1
        assert extended_header_3['receiver_point'] == 1
        assert extended_header_3['receiver_point_index'] == 1

    def test_bad_alias_filter(self):
        """
        Tests for when alias filter is written as an int32 rather than BSD in
        the channel descriptor block.
        """
        with open(self.bad_alias_filter, 'rb') as fi:
            out = rc._read_channel_set(fi, 0)
        alias_freq = out['alias_filter_frequency']
        assert alias_freq == 207

    def test_3_channel_header(self):
        """
        Tests for header which has three traces but identical RU channel
        number.
        """
        with open(self.header_block_same_ru_code, 'rb') as fi:
            num_records = rc._cmp_nbr_records(fi)
        assert num_records == 2180 * 3



# -*- coding: utf-8 -*-
"""
The obspy.clients.seedlink.client.slstate test suite.
"""

import pytest

from obspy.clients.seedlink.client.slstate import SLState


pytestmark = pytest.mark.network


class TestSLState():

    def test_issue561(self):
        """
        Assure that different state objects don't share data buffers.
        """
        slstate1 = SLState()
        slstate2 = SLState()

        assert id(slstate1.databuf) != id(slstate2.databuf)
        assert id(slstate1.packed_buf) != id(slstate2.packed_buf)



# -*- coding: utf-8 -*-
"""
obspy.io.nied.fnet - F-net moment tensor file read support for ObsPy
====================================================================
F-net moment tensor file format support for ObsPy.

:copyright:
    The ObsPy Development Team (devs@obspy.org) and Yannik Behr
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import re
import uuid

from obspy import UTCDateTime
from obspy.core.event import (Catalog, Comment, Event,
                              Origin, Magnitude, FocalMechanism, MomentTensor,
                              Tensor, NodalPlane, NodalPlanes)
from . import util


class FNETMTException(Exception):
    pass


def _get_resource_id(name, res_type, tag=None):
    """
    Helper function to create consistent resource ids.
    """
    res_id = "smi:local/fnetmt/%s/%s" % (name, res_type)
    if tag is not None:
        res_id += "#" + tag
    return res_id


def _buffer_proxy(filename_or_buf, function, reset_fp=True,
                  file_mode="rb", *args, **kwargs):
    """
    Calls a function with an open file or file-like object as the first
    argument. If the file originally was a filename, the file will be
    opened, otherwise it will just be passed to the underlying function.

    :param filename_or_buf: File to pass.
    :type filename_or_buf: str, open file, or file-like object
    :param function: The function to call.
    :param reset_fp: If True, the file pointer will be set to the initial
        position after the function has been called.
    :type reset_fp: bool
    :param file_mode: Mode to open file in if necessary.
    """
    try:
        position = filename_or_buf.tell()
        is_buffer = True
    except AttributeError:
        is_buffer = False

    if is_buffer is True:
        ret_val = function(filename_or_buf, *args, **kwargs)
        if reset_fp:
            filename_or_buf.seek(position, 0)
        return ret_val
    else:
        with open(filename_or_buf, file_mode) as fh:
            return function(fh, *args, **kwargs)


def _is_fnetmt_catalog(filename_or_buf):
    """
    Checks if the file is an F-net moment tensor file.

    :param filename_or_buf: File to test.
    :type filename_or_buf: str or file-like object
    """
    try:
        return _buffer_proxy(filename_or_buf, _internal_is_fnetmt_catalog,
                             reset_fp=True)
    # Happens for example when passing the data as a string which would be
    # interpreted as a filename.
    except (OSError):
        return False


def _internal_is_fnetmt_catalog(buf):
    """
    Test whether file is an F-net moment tensor catalog file by reading the
    header and the first data line. Reads at most 40 lines.

    :param buf: File to read.
    :type buf: open file or file-like object
    """
    cnt = 0
    try:
        while True:
            line = buf.readline()
            if not line:
                return False
            line = line.decode()
            # read at most 40 lines
            if cnt > 40:
                return False
            if line.find('Total Number') != -1:
                match = re.search(r'Total Number:\s+(\d+)\s+', line)
                if match:
                    nevents = int(match.group(1))
            if line.startswith('Origin Time'):
                if nevents > 0:
                    data = buf.readline()
                    a = data.split()
                    if len(a) != 21:
                        return False
                    return True
            cnt += 1
    except Exception:
        return False
    else:
        return True


def _read_fnetmt_catalog(filename_or_buf, **kwargs):
    """
    Reads an F-net moment tensor catalog file to a
    :class:`~obspy.core.event.Catalog` object.

    :param filename_or_buf: File to read.
    :type filename_or_buf: str or file-like object
    """
    return _buffer_proxy(filename_or_buf, _internal_read_fnetmt_catalog,
                         **kwargs)


def _internal_read_fnetmt_catalog(buf, **kwargs):
    """
    Reads an F-net moment tensor catalog file to a
    :class:`~obspy.core.event.Catalog` object.

    :param buf: File to read.
    :type buf: open file or file-like object
    """
    events = []
    cur_pos = buf.tell()

    # This also works with BytesIO and what not.
    buf.seek(0, 2)
    size = buf.tell()
    buf.seek(cur_pos, 0)

    # First read the headerlines containing the data request parameters
    headerlines = []
    while buf.tell() < size:
        line = buf.readline().decode()
        if line.find('Total Number') != -1:
            match = re.search(r'Total Number:\s+(\d+)\s+', line)
            if match:
                nevents = int(match.group(1))
        if line.startswith('Origin Time'):
            break
        headerlines.append(line)
    cur_pos = buf.tell()

    # Now read the catalog
    while buf.tell() < size:
        line = buf.readline().strip()
        # If there is something, jump back to the beginning of the line and
        # read the next event.
        if line:
            events.append(_internal_read_single_fnetmt_entry(line.decode()))

    # Consistency check
    if len(events) != nevents:
        raise FNETMTException('Parsing failed! Expected %d events but read %d.'
                              % (nevents, len(events)))

    return Catalog(resource_id=_get_resource_id("catalog", str(uuid.uuid4())),
                   events=events, description=headerlines[:-1])


def _internal_read_single_fnetmt_entry(line, **kwargs):
    """
    Reads a single F-net moment tensor solution to a
    :class:`~obspy.core.event.event.Event` object.

    :param line: String containing moment tensor information.
    :type line: str
    """

    a = line.split()
    try:
        ot = UTCDateTime.strptime(a[0], '%Y/%m/%d,%H:%M:%S.%f')
    except ValueError:
        ot = UTCDateTime.strptime(a[0], '%Y/%m/%d,%H:%M:%S')
    lat, lon, depjma, magjma = map(float, a[1:5])
    depjma *= 1000
    region = a[5]
    strike = tuple(map(int, a[6].split(';')))
    dip = tuple(map(int, a[7].split(';')))
    rake = tuple(map(int, a[8].split(';')))
    mo = float(a[9])
    depmt = float(a[10]) * 1000
    magmt = float(a[11])
    var_red = float(a[12])
    mxx, mxy, mxz, myy, myz, mzz, unit = map(float, a[13:20])

    event_name = util.gen_sc3_id(ot)
    e = Event(event_type="earthquake")
    e.resource_id = _get_resource_id(event_name, 'event')

    # Standard JMA solution
    o_jma = Origin(time=ot, latitude=lat, longitude=lon,
                   depth=depjma, depth_type="from location",
                   region=region)
    o_jma.resource_id = _get_resource_id(event_name,
                                         'origin', 'JMA')
    m_jma = Magnitude(mag=magjma, magnitude_type='ML',
                      origin_id=o_jma.resource_id)
    m_jma.resource_id = _get_resource_id(event_name,
                                         'magnitude', 'JMA')
    # MT solution
    o_mt = Origin(time=ot, latitude=lat, longitude=lon,
                  depth=depmt, region=region,
                  depth_type="from moment tensor inversion")
    o_mt.resource_id = _get_resource_id(event_name,
                                        'origin', 'MT')
    m_mt = Magnitude(mag=magmt, magnitude_type='Mw',
                     origin_id=o_mt.resource_id)
    m_mt.resource_id = _get_resource_id(event_name,
                                        'magnitude', 'MT')
    foc_mec = FocalMechanism(triggering_origin_id=o_jma.resource_id)
    foc_mec.resource_id = _get_resource_id(event_name,
                                           "focal_mechanism")
    nod1 = NodalPlane(strike=strike[0], dip=dip[0], rake=rake[0])
    nod2 = NodalPlane(strike=strike[1], dip=dip[1], rake=rake[1])
    nod = NodalPlanes(nodal_plane_1=nod1, nodal_plane_2=nod2)
    foc_mec.nodal_planes = nod

    tensor = Tensor(m_rr=mxx, m_tt=myy, m_pp=mzz, m_rt=mxy, m_rp=mxz, m_tp=myz)
    cm = Comment(text="Basis system: North,East,Down (Jost and \
    Herrmann 1989")
    cm.resource_id = _get_resource_id(event_name, 'comment', 'mt')
    mt = MomentTensor(derived_origin_id=o_mt.resource_id,
                      moment_magnitude_id=m_mt.resource_id,
                      scalar_moment=mo, comments=[cm],
                      tensor=tensor, variance_reduction=var_red)
    mt.resource_id = _get_resource_id(event_name,
                                      'moment_tensor')
    foc_mec.moment_tensor = mt
    e.origins = [o_jma, o_mt]
    e.magnitudes = [m_jma, m_mt]
    e.focal_mechanisms = [foc_mec]
    e.preferred_magnitude_id = m_mt.resource_id.id
    e.preferred_origin_id = o_mt.resource_id.id
    e.preferred_focal_mechanism_id = foc_mec.resource_id.id

    e.scope_resource_ids()

    return e



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module to create and use a connection to a SeedLink server using a
SeedLinkConnection object.

A new SeedLink application can be created by sub-classing SLClient and
overriding at least the packet_handler method of SLClient.

Part of Python implementation of libslink of Chad Trabant and
JSeedLink of Anthony Lomax

:copyright:
    The ObsPy Development Team (devs@obspy.org) & Anthony Lomax
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import sys

from obspy.core.util.decorator import deprecated_keywords
from .client.seedlinkconnection import SeedLinkConnection
from .seedlinkexception import SeedLinkException
from .slpacket import SLPacket


USAGE = """
## General program options ##
-V             report program version
-h             show this usage message
-v             be more verbose, multiple flags can be used
-p             print details of data packets
-nd delay      network re-connect delay (seconds), default 30
-nt timeout    network timeout (seconds), re-establish connection if no
               data/keepalives are received in this time, default 600
-k interval    send keepalive (heartbeat) packets this often (seconds)
-x statefile   save/restore stream state information to this file
-t begintime   sets a beginning time for the initiation of data transmission
               (year,month,day,hour,minute,second)
-e endtime     sets an end time for windowed data transmission
               (year,month,day,hour,minute,second)
-i infolevel   request this INFO level, write response to std out, and exit
               infolevel is one of: ID, STATIONS, STREAMS, GAPS, CONNECTIONS,
               ALL

## Data stream selection ##
-l listfile    read a stream list from this file for multi-station mode
-s selectors   selectors for uni-station or default for multi-station
-S streams     select streams for multi-station (requires SeedLink >= 2.5)
  'streams' = 'stream1[:selectors1],stream2[:selectors2],...'
       'stream' is in NET_STA format, for example:
       -S \"IU_KONO:BHE BHN,GE_WLF,MN_AQU:HH?.D\"

<[host]:port>  Address of the SeedLink server in host:port format
               if host is omitted (i.e. ':18000'), localhost is assumed
"""


class SLClient(object):
    """
    Basic class to create and use a connection to a SeedLink server using a
    SeedLinkConnection object.

    A new SeedLink application can be created by sub-classing SLClient and
    overriding at least the packet_handler method of SLClient.

    :var slconn: SeedLinkConnection object for communicating with the
        SeedLinkConnection over a socket.
    :type slconn: SeedLinkConnection
    :var verbose: Verbosity level, 0 is lowest.
    :type verbose: int
    :var ppackets: Flag to indicate show detailed packet information.
    :type  ppackets: bool
    :var streamfile: Name of file containing stream list for multi-station
        mode.
    :type  streamfile: str
    :var selectors: Selectors for uni-station or default selectors for
        multi-station.
    :type  selectors: str
    :var multiselect: Selectors for multi-station.
    :type  multiselect: str
    :var statefile: Name of file for reading (if exists) and storing state.
    :type  statefile: str
    :var begin_time: Beginning of time window for read start in past.
    :type  begin_time: str
    :var end_time: End of time window for reading windowed data.
    :type  end_time: str
    :var infolevel: INFO LEVEL for info request only.
    :type  infolevel: str
    :type timeout: float
    :param timeout: Timeout in seconds, passed on to the underlying
        SeedLinkConnection.
    """
    VERSION = "1.2.0X00"
    VERSION_YEAR = "2011"
    VERSION_DATE = "24Nov" + VERSION_YEAR
    COPYRIGHT_YEAR = VERSION_YEAR
    PROGRAM_NAME = "SLClient v" + VERSION
    VERSION_INFO = PROGRAM_NAME + " (" + VERSION_DATE + ")"

    @deprecated_keywords({"loglevel": None})
    def __init__(self, loglevel=None, timeout=None):
        """
        Creates a new instance of SLClient with the specified logging object
        """
        self.verbose = 0
        self.ppackets = False
        self.streamfile = None
        self.selectors = None
        self.multiselect = None
        self.statefile = None
        self.begin_time = None
        self.end_time = None
        self.infolevel = None
        self.timeout = timeout
        self.slconn = SeedLinkConnection(timeout=timeout)

    def parse_cmd_line_args(self, args):
        """
        Parses the command line arguments.

        :type args: list
        :param args: main method arguments.
        :return: -1 on error, 1 if version or help argument found, 0 otherwise.
        """
        if len(args) < 2:
            self.print_usage(False)
            return 1
        optind = 1
        while optind < len(args):
            if args[optind] == "-V":
                print(self.VERSION_INFO, file=sys.stderr)
                return 1
            elif args[optind] == "-h":
                self.print_usage(False)
                return 1
            elif args[optind].startswith("-v"):
                self.verbose += len(args[optind]) - 1
            elif args[optind] == "-p":
                self.ppackets = True
            elif args[optind] == "-nt":
                optind += 1
                self.slconn.set_net_timeout(int(args[optind]))
            elif args[optind] == "-nd":
                optind += 1
                self.slconn.set_net_delay(int(args[optind]))
            elif args[optind] == "-k":
                optind += 1
                self.slconn.set_keep_alive(int(args[optind]))
            elif args[optind] == "-l":
                optind += 1
                self.streamfile = args[optind]
            elif args[optind] == "-s":
                optind += 1
                self.selectors = args[optind]
            elif args[optind] == "-S":
                optind += 1
                self.multiselect = args[optind]
            elif args[optind] == "-x":
                optind += 1
                self.statefile = args[optind]
            elif args[optind] == "-t":
                optind += 1
                self.begin_time = args[optind]
            elif args[optind] == "-e":
                optind += 1
                self.end_time = args[optind]
            elif args[optind] == "-i":
                optind += 1
                self.infolevel = args[optind]
            elif args[optind].startswith("-"):
                print("Unknown option: " + args[optind], file=sys.stderr)
                return -1
            elif self.slconn.get_sl_address() is None:
                self.slconn.set_sl_address(args[optind])
            else:
                print("Unknown option: " + args[optind], file=sys.stderr)
                return -1
            optind += 1
        return 0

    def initialize(self):
        """
        Initializes this SLClient.
        """
        if self.slconn.get_sl_address() is None:
            message = "no SeedLink server specified"
            raise SeedLinkException(message)

        if self.verbose >= 2:
            self.ppackets = True
        if self.slconn.get_sl_address().startswith(":"):
            self.slconn.set_sl_address("127.0.0.1" +
                                       self.slconn.get_sl_address())
        if self.streamfile is not None:
            self.slconn.read_stream_list(self.streamfile, self.selectors)
        if self.multiselect is not None:
            self.slconn.parse_stream_list(self.multiselect, self.selectors)
        else:
            if self.streamfile is None:
                self.slconn.set_uni_params(self.selectors, -1, None)
        if self.statefile is not None:
            self.slconn.set_state_file(self.statefile)
        else:
            if self.begin_time is not None:
                self.slconn.set_begin_time(self.begin_time)
            if self.end_time is not None:
                self.slconn.set_end_time(self.end_time)

    def run(self, packet_handler=None):
        """
        Start this SLClient.

        :type packet_handler: callable
        :param packet_handler: Custom packet handler funtion to override
            `self.packet_handler` for this seedlink request. The function will
            be repeatedly called with two arguments: the current packet counter
            (`int`) and the currently served seedlink packet
            (:class:`~obspy.clients.seedlink.slclient.SLPacket`). The
            function should return `True` to abort the request or `False` to
             continue the request.
        """
        if packet_handler is None:
            packet_handler = self.packet_handler
        if self.infolevel is not None:
            self.slconn.request_info(self.infolevel)
        # Loop with the connection manager
        count = 1
        slpack = self.slconn.collect()
        while slpack is not None:
            if (slpack == SLPacket.SLTERMINATE):
                break
            try:
                # do something with packet
                terminate = packet_handler(count, slpack)
                if terminate:
                    break
            except SeedLinkException as sle:
                print(self.__class__.__name__ + ": " + sle.value)
            if count >= sys.maxsize:
                count = 1
                print("DEBUG INFO: " + self.__class__.__name__ + ":", end=' ')
                print("Packet count reset to 1")
            else:
                count += 1
            slpack = self.slconn.collect()

        # Close the SeedLinkConnection
        self.slconn.close()

    def packet_handler(self, count, slpack):
        """
        Processes each packet received from the SeedLinkConnection.

        This method should be overridden when sub-classing SLClient.

        :type count: int
        :param count:  Packet counter.
        :type slpack: :class:`~obspy.clients.seedlink.slpacket.SLPacket`
        :param slpack: packet to process.

        :rtype: bool
        :return: True if connection to SeedLink server should be closed and
            session terminated, False otherwise.
        """
        # check if not a complete packet
        if slpack is None or (slpack == SLPacket.SLNOPACKET) or \
                (slpack == SLPacket.SLERROR):
            return False

        # get basic packet info
        seqnum = slpack.get_sequence_number()
        type = slpack.get_type()

        # process INFO packets here
        if (type == SLPacket.TYPE_SLINF):
            return False
        if (type == SLPacket.TYPE_SLINFT):
            print("Complete INFO:\n" + self.slconn.get_info_string())
            if self.infolevel is not None:
                return True
            else:
                return False

        # can send an in-line INFO request here
        try:
            # if (count % 100 == 0 and not self.slconn.state.expect_info):
            if (count % 100 == 0):
                infostr = "ID"
                self.slconn.request_info(infostr)
        except SeedLinkException as sle:
            print(self.__class__.__name__ + ": " + sle.value)

        # if here, must be a data blockette
        print(self.__class__.__name__ + ": packet seqnum:", end=' ')
        print(str(seqnum) + ": blockette type: " + str(type))
        if not self.ppackets:
            return False

        # process packet data
        trace = slpack.get_trace()
        if trace is not None:
            print(self.__class__.__name__ + ": blockette contains a trace: ")
            print(trace.id, trace.stats['starttime'], end=' ')
            print(" dt:" + str(1.0 / trace.stats['sampling_rate']), end=' ')
            print(" npts:" + str(trace.stats['npts']), end=' ')
            print(" sampletype:" + str(trace.stats['sampletype']), end=' ')
            print(" dataquality:" + str(trace.stats['dataquality']))
            if self.verbose >= 3:
                print(self.__class__.__name__ + ":")
                print("blockette contains a trace: " + str(trace.stats))
        else:
            print(self.__class__.__name__ + ": blockette contains no trace")
        return False

    def print_usage(self, concise=True):
        """
        Prints the usage message for this class.
        """
        print("\nUsage: python %s [options] <[host]:port>" %
              (self.__class__.__name__))
        if concise:
            usage = "Use '-h' for detailed help"
        else:
            usage = USAGE
        print(usage)

    @classmethod
    def main(cls, args):
        """
        Main method - creates and runs an SLClient using the specified
        command line arguments
        """
        sl_client = SLClient()
        rval = sl_client.parse_cmd_line_args(args)
        if (rval != 0):
            sys.exit(rval)
        sl_client.initialize()
        sl_client.run()


if __name__ == '__main__':
    SLClient.main(sys.argv)



import matplotlib.pyplot as plt

import obspy
from obspy.clients.fdsn import Client
from obspy.signal.trigger import recursive_sta_lta, trigger_onset


# Retrieve waveforms via FDSNWS
client = Client("LMU")
t = obspy.UTCDateTime("2009-08-24 00:19:45")
st = client.get_waveforms('BW', 'RTSH', '', 'EHZ', t, t + 50)

# For convenience
tr = st[0]  # only one trace in mseed volume
df = tr.stats.sampling_rate

# Characteristic function and trigger onsets
cft = recursive_sta_lta(tr.data, int(2.5 * df), int(10. * df))
on_of = trigger_onset(cft, 3.5, 0.5)

# Plotting the results
ax = plt.subplot(211)
plt.plot(tr.data, 'k')
ymin, ymax = ax.get_ylim()
plt.vlines(on_of[:, 0], ymin, ymax, color='r', linewidth=2)
plt.vlines(on_of[:, 1], ymin, ymax, color='b', linewidth=2)
plt.subplot(212, sharex=ax)
plt.plot(cft, 'k')
plt.hlines([3.5, 0.5], 0, len(cft), color=['r', 'b'], linestyle='--')
plt.axis('tight')
plt.show()



# -*- coding: utf-8 -*-
import socket


def main():
    """
    An example test for seed link
    """
    HOST = 'geofon.gfz.de'    # The remote host
    PORT = 18000              # The remote port
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    print('s.connect((', HOST, PORT, '))')
    s.connect((HOST, PORT))
    sendbytes = b'HELLO\r'
    print('Sent:', repr(sendbytes))
    s.send(sendbytes)
    data = s.recv(1024)
    print('Received:', repr(data))
    s.close()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Keyhole Markup Language (KML) output support in ObsPy

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from math import log

from lxml.etree import Element, SubElement, tostring
from matplotlib.pyplot import get_cmap

from obspy import UTCDateTime
from obspy.core.event import Catalog
from obspy.core.inventory.inventory import Inventory


def inventory_to_kml_string(
        inventory,
        icon_url="https://maps.google.com/mapfiles/kml/shapes/triangle.png",
        icon_size=1.5, label_size=1.0, cmap="Paired", encoding="UTF-8",
        timespans=True, strip_far_future_end_times=True):
    """
    Convert an :class:`~obspy.core.inventory.inventory.Inventory` to a KML
    string representation.

    :type inventory: :class:`~obspy.core.inventory.inventory.Inventory`
    :param inventory: Input station metadata.
    :type icon_url: str
    :param icon_url: Internet URL of icon to use for station (e.g. PNG image).
    :type icon_size: float
    :param icon_size: Icon size.
    :type label_size: float
    :param label_size: Label size.
    :type encoding: str
    :param encoding: Encoding used for XML string.
    :type timespans: bool
    :param timespans: Whether to add timespan information to the single station
        elements in the KML or not. If timespans are used, the displayed
        information in e.g. Google Earth will represent a snapshot in time,
        such that using the time slider different states of the inventory in
        time can be visualized. If timespans are not used, any station active
        at any point in time is always shown.
    :type strip_far_future_end_times: bool
    :param strip_far_future_end_times: Leave out likely fictitious end times of
        stations (more than twenty years after current time). Far future end
        times may produce time sliders with bad overall time span in third
        party applications viewing the KML file.
    :rtype: bytes
    :return: Encoded byte string containing KML information of the station
        metadata.
    """
    twenty_years_from_now = UTCDateTime() + 3600 * 24 * 365 * 20
    # construct the KML file
    kml = Element("kml")
    kml.set("xmlns", "http://www.opengis.net/kml/2.2")

    document = SubElement(kml, "Document")
    SubElement(document, "name").text = "Inventory"

    # style definition
    cmap = get_cmap(name=cmap, lut=len(inventory.networks))
    for i in range(len(inventory.networks)):
        color = _rgba_tuple_to_kml_color_code(cmap(i))
        style = SubElement(document, "Style")
        style.set("id", "station_%i" % i)

        iconstyle = SubElement(style, "IconStyle")
        SubElement(iconstyle, "color").text = color
        SubElement(iconstyle, "scale").text = str(icon_size)
        icon = SubElement(iconstyle, "Icon")
        SubElement(icon, "href").text = icon_url
        hotspot = SubElement(iconstyle, "hotSpot")
        hotspot.set("x", "0.5")
        hotspot.set("y", "0.5")
        hotspot.set("xunits", "fraction")
        hotspot.set("yunits", "fraction")

        labelstyle = SubElement(style, "LabelStyle")
        SubElement(labelstyle, "color").text = color
        SubElement(labelstyle, "scale").text = str(label_size)

    for i, net in enumerate(inventory):
        folder = SubElement(document, "Folder")
        SubElement(folder, "name").text = str(net.code)
        SubElement(folder, "open").text = "1"

        SubElement(folder, "description").text = str(net)

        style = SubElement(folder, "Style")
        liststyle = SubElement(style, "ListStyle")
        SubElement(liststyle, "listItemType").text = "check"
        SubElement(liststyle, "bgColor").text = "00ffff"
        SubElement(liststyle, "maxSnippetLines").text = "5"

        # add one marker per station code
        for sta in net:
            placemark = SubElement(folder, "Placemark")
            SubElement(placemark, "name").text = ".".join((net.code, sta.code))
            SubElement(placemark, "styleUrl").text = "#station_%i" % i
            SubElement(placemark, "color").text = color
            if sta.longitude is not None and sta.latitude is not None:
                point = SubElement(placemark, "Point")
                SubElement(point, "coordinates").text = "%.6f,%.6f,0" % \
                    (sta.longitude, sta.latitude)

            SubElement(placemark, "description").text = str(sta)

            if timespans:
                start = sta.start_date
                end = sta.end_date
                if start is not None or end is not None:
                    timespan = SubElement(placemark, "TimeSpan")
                    if start is not None:
                        SubElement(timespan, "begin").text = str(start)
                    if end is not None:
                        if not strip_far_future_end_times or \
                                end < twenty_years_from_now:
                            SubElement(timespan, "end").text = str(end)
        if timespans:
            start = net.start_date
            end = net.end_date
            if start is not None or end is not None:
                timespan = SubElement(folder, "TimeSpan")
                if start is not None:
                    SubElement(timespan, "begin").text = str(start)
                if end is not None:
                    if not strip_far_future_end_times or \
                            end < twenty_years_from_now:
                        SubElement(timespan, "end").text = str(end)

    # generate and return KML string
    return tostring(kml, pretty_print=True, xml_declaration=True,
                    encoding=encoding)


def catalog_to_kml_string(
        catalog,
        icon_url="https://maps.google.com/mapfiles/kml/shapes/earthquake.png",
        label_func=None, icon_size_func=None, encoding="UTF-8",
        timestamps=True):
    """
    Convert a :class:`~obspy.core.event.Catalog` to a KML string
    representation.

    :type catalog: :class:`~obspy.core.event.Catalog`
    :param catalog: Input catalog data.
    :type icon_url: str
    :param icon_url: Internet URL of icon to use for events (e.g. PNG image).
    :type label_func: func
    :type label_func: Custom function to use for determining each event's
        label. User provided function is supposed to take an
        :class:`~obspy.core.event.event.Event` object as single argument, e.g.
        for empty labels use `label_func=lambda x: ""`.
    :type icon_size_func: func
    :type icon_size_func: Custom function to use for determining each
        event's icon size. User provided function should take an
        :class:`~obspy.core.event.event.Event` object as single argument
        and return a float.
    :type encoding: str
    :param encoding: Encoding used for XML string.
    :type timestamps: bool
    :param timestamps: Whether to add timestamp information to the event
        elements in the KML or not. If timestamps are used, the displayed
        information in e.g. Google Earth will represent a snapshot in time,
        such that using the time slider different states of the catalog in time
        can be visualized. If timespans are not used, any event happening at
        any point in time is always shown.
    :rtype: bytes
    :return: Encoded byte string containing KML information of the event
        metadata.
    """
    # default label and size functions
    if not label_func:
        def label_func(event):
            origin = (event.preferred_origin() or
                      event.origins and event.origins[0] or
                      None)
            mag = (event.preferred_magnitude() or
                   event.magnitudes and event.magnitudes[0] or
                   None)
            label = origin.time and str(origin.time.date) or ""
            if mag:
                label += " %.1f" % mag.mag
            return label
    if not icon_size_func:
        def icon_size_func(event):
            mag = (event.preferred_magnitude() or
                   event.magnitudes and event.magnitudes[0] or
                   None)
            if mag:
                try:
                    icon_size = 1.2 * log(1.5 + mag.mag)
                except ValueError:
                    icon_size = 0.1
            else:
                icon_size = 0.5
            return icon_size

    # construct the KML file
    kml = Element("kml")
    kml.set("xmlns", "http://www.opengis.net/kml/2.2")

    document = SubElement(kml, "Document")
    SubElement(document, "name").text = "Catalog"

    # style definitions for earthquakes
    style = SubElement(document, "Style")
    style.set("id", "earthquake")

    iconstyle = SubElement(style, "IconStyle")
    SubElement(iconstyle, "scale").text = "0.5"
    icon = SubElement(iconstyle, "Icon")
    SubElement(icon, "href").text = icon_url
    hotspot = SubElement(iconstyle, "hotSpot")
    hotspot.set("x", "0.5")
    hotspot.set("y", "0.5")
    hotspot.set("xunits", "fraction")
    hotspot.set("yunits", "fraction")

    labelstyle = SubElement(style, "LabelStyle")
    SubElement(labelstyle, "color").text = "ff0000ff"
    SubElement(labelstyle, "scale").text = "0.8"

    folder = SubElement(document, "Folder")
    SubElement(folder, "name").text = "Catalog"
    SubElement(folder, "open").text = "1"

    SubElement(folder, "description").text = str(catalog)

    style = SubElement(folder, "Style")
    liststyle = SubElement(style, "ListStyle")
    SubElement(liststyle, "listItemType").text = "check"
    SubElement(liststyle, "bgColor").text = "00ffffff"
    SubElement(liststyle, "maxSnippetLines").text = "5"

    # add one marker per event
    for event in catalog:
        origin = (event.preferred_origin() or
                  event.origins and event.origins[0] or
                  None)

        placemark = SubElement(folder, "Placemark")
        SubElement(placemark, "name").text = label_func(event)
        SubElement(placemark, "styleUrl").text = "#earthquake"
        style = SubElement(placemark, "Style")
        icon_style = SubElement(style, "IconStyle")
        liststyle = SubElement(style, "ListStyle")
        SubElement(liststyle, "maxSnippetLines").text = "5"
        SubElement(icon_style, "scale").text = "%.5f" % icon_size_func(event)
        if origin:
            if origin.longitude is not None and origin.latitude is not None:
                point = SubElement(placemark, "Point")
                SubElement(point, "coordinates").text = "%.6f,%.6f,0" % \
                    (origin.longitude, origin.latitude)

        SubElement(placemark, "description").text = str(event)

        if timestamps:
            time = _get_event_timestamp(event)
            if time is not None:
                SubElement(placemark, "TimeStamp").text = str(time)

    # generate and return KML string
    return tostring(kml, pretty_print=True, xml_declaration=True,
                    encoding=encoding)


def _write_kml(obj, filename, **kwargs):
    """
    Write :class:`~obspy.core.inventory.inventory.Inventory` or
    :class:`~obspy.core.event.Catalog` object to a KML file.
    For additional parameters see :meth:`inventory_to_kml_string` and
    :meth:`catalog_to_kml_string`.

    :type obj: :class:`~obspy.core.event.Catalog` or
        :class:`~obspy.core.inventory.inventory.Inventory`
    :param obj: ObsPy object for KML output
    :type filename: str
    :param filename: Filename to write to. Suffix ".kml" will be appended if
        not already present.
    """
    if isinstance(obj, Catalog):
        kml_string = catalog_to_kml_string(obj, **kwargs)
    elif isinstance(obj, Inventory):
        kml_string = inventory_to_kml_string(obj, **kwargs)
    else:
        msg = ("Object for KML output must be "
               "a Catalog or Inventory.")
        raise TypeError(msg)
    if not filename.endswith(".kml"):
        filename += ".kml"
    with open(filename, "wb") as fh:
        fh.write(kml_string)


def _rgba_tuple_to_kml_color_code(rgba):
    """
    Convert tuple of (red, green, blue, alpha) float values (0.0-1.0) to KML
    hex color code string "aabbggrr".
    """
    try:
        r, g, b, a = rgba
    except Exception:
        r, g, b = rgba
        a = 1.0
    return "".join(["%02x" % int(x * 255) for x in (a, b, g, r)])


def _get_event_timestamp(event):
    """
    Get timestamp information for the event. Search is perfomed in the
    following order:

     - origin time of preferred origin
     - origin time of first origin found that has a origin time
     - minimum of all found pick times
     - `None` if no time is found in the above search
    """
    origin = event.preferred_origin()
    if origin is not None and origin.time is not None:
        return origin.time
    for origin in event.origins:
        if origin.time is not None:
            return origin.time
    pick_times = [pick.time for pick in event.picks
                  if pick.time is not None]
    if pick_times:
        return min(pick_times)
    return None


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Get the Flinn-Engdahl region name from longitude and latitude.
"""
from argparse import ArgumentParser

from obspy import __version__
from obspy.geodetics import FlinnEngdahl


def main(argv=None):
    parser = ArgumentParser(prog='obspy-flinn-engdahl',
                            description=__doc__.strip())
    parser.add_argument('-V', '--version', action='version',
                        version='%(prog)s ' + __version__)
    parser.add_argument('longitude', type=float,
                        help='Longitude (in degrees) of point. Positive for '
                             'East, negative for West.')
    parser.add_argument('latitude', type=float,
                        help='Latitude (in degrees) of point. Positive for '
                             'North, negative for South.')
    args = parser.parse_args(argv)

    flinn_engdahl = FlinnEngdahl()
    print(flinn_engdahl.get_region(args.longitude, args.latitude))


if __name__ == '__main__':
    # It is not possible to put the code of main directly here.
    # This script is automatically installed with name obspy-... by
    # setup.py to the Scripts or bin directory of your Python distribution.
    # setup.py needs a function to which its scripts can be linked.
    main()
